from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class InvalidAuthorityJSON(typing.TypedDict):
    kind: typing.Literal["InvalidAuthority"]


class InvalidTokenMinterStateJSON(typing.TypedDict):
    kind: typing.Literal["InvalidTokenMinterState"]


class ProgramPausedJSON(typing.TypedDict):
    kind: typing.Literal["ProgramPaused"]


class InvalidTokenPairStateJSON(typing.TypedDict):
    kind: typing.Literal["InvalidTokenPairState"]


class InvalidLocalTokenStateJSON(typing.TypedDict):
    kind: typing.Literal["InvalidLocalTokenState"]


class InvalidPauserJSON(typing.TypedDict):
    kind: typing.Literal["InvalidPauser"]


class InvalidTokenControllerJSON(typing.TypedDict):
    kind: typing.Literal["InvalidTokenController"]


class BurnAmountExceededJSON(typing.TypedDict):
    kind: typing.Literal["BurnAmountExceeded"]


class InvalidAmountJSON(typing.TypedDict):
    kind: typing.Literal["InvalidAmount"]


@dataclass
class InvalidAuthority:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "InvalidAuthority"

    @classmethod
    def to_json(cls) -> InvalidAuthorityJSON:
        return InvalidAuthorityJSON(
            kind="InvalidAuthority",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidAuthority": {},
        }


@dataclass
class InvalidTokenMinterState:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "InvalidTokenMinterState"

    @classmethod
    def to_json(cls) -> InvalidTokenMinterStateJSON:
        return InvalidTokenMinterStateJSON(
            kind="InvalidTokenMinterState",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidTokenMinterState": {},
        }


@dataclass
class ProgramPaused:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "ProgramPaused"

    @classmethod
    def to_json(cls) -> ProgramPausedJSON:
        return ProgramPausedJSON(
            kind="ProgramPaused",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ProgramPaused": {},
        }


@dataclass
class InvalidTokenPairState:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "InvalidTokenPairState"

    @classmethod
    def to_json(cls) -> InvalidTokenPairStateJSON:
        return InvalidTokenPairStateJSON(
            kind="InvalidTokenPairState",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidTokenPairState": {},
        }


@dataclass
class InvalidLocalTokenState:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "InvalidLocalTokenState"

    @classmethod
    def to_json(cls) -> InvalidLocalTokenStateJSON:
        return InvalidLocalTokenStateJSON(
            kind="InvalidLocalTokenState",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidLocalTokenState": {},
        }


@dataclass
class InvalidPauser:
    discriminator: typing.ClassVar = 5
    kind: typing.ClassVar = "InvalidPauser"

    @classmethod
    def to_json(cls) -> InvalidPauserJSON:
        return InvalidPauserJSON(
            kind="InvalidPauser",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidPauser": {},
        }


@dataclass
class InvalidTokenController:
    discriminator: typing.ClassVar = 6
    kind: typing.ClassVar = "InvalidTokenController"

    @classmethod
    def to_json(cls) -> InvalidTokenControllerJSON:
        return InvalidTokenControllerJSON(
            kind="InvalidTokenController",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidTokenController": {},
        }


@dataclass
class BurnAmountExceeded:
    discriminator: typing.ClassVar = 7
    kind: typing.ClassVar = "BurnAmountExceeded"

    @classmethod
    def to_json(cls) -> BurnAmountExceededJSON:
        return BurnAmountExceededJSON(
            kind="BurnAmountExceeded",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "BurnAmountExceeded": {},
        }


@dataclass
class InvalidAmount:
    discriminator: typing.ClassVar = 8
    kind: typing.ClassVar = "InvalidAmount"

    @classmethod
    def to_json(cls) -> InvalidAmountJSON:
        return InvalidAmountJSON(
            kind="InvalidAmount",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "InvalidAmount": {},
        }


TokenMinterErrorKind = typing.Union[
    InvalidAuthority,
    InvalidTokenMinterState,
    ProgramPaused,
    InvalidTokenPairState,
    InvalidLocalTokenState,
    InvalidPauser,
    InvalidTokenController,
    BurnAmountExceeded,
    InvalidAmount,
]
TokenMinterErrorJSON = typing.Union[
    InvalidAuthorityJSON,
    InvalidTokenMinterStateJSON,
    ProgramPausedJSON,
    InvalidTokenPairStateJSON,
    InvalidLocalTokenStateJSON,
    InvalidPauserJSON,
    InvalidTokenControllerJSON,
    BurnAmountExceededJSON,
    InvalidAmountJSON,
]


def from_decoded(obj: dict) -> TokenMinterErrorKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "InvalidAuthority" in obj:
        return InvalidAuthority()
    if "InvalidTokenMinterState" in obj:
        return InvalidTokenMinterState()
    if "ProgramPaused" in obj:
        return ProgramPaused()
    if "InvalidTokenPairState" in obj:
        return InvalidTokenPairState()
    if "InvalidLocalTokenState" in obj:
        return InvalidLocalTokenState()
    if "InvalidPauser" in obj:
        return InvalidPauser()
    if "InvalidTokenController" in obj:
        return InvalidTokenController()
    if "BurnAmountExceeded" in obj:
        return BurnAmountExceeded()
    if "InvalidAmount" in obj:
        return InvalidAmount()
    raise ValueError("Invalid enum object")


def from_json(obj: TokenMinterErrorJSON) -> TokenMinterErrorKind:
    if obj["kind"] == "InvalidAuthority":
        return InvalidAuthority()
    if obj["kind"] == "InvalidTokenMinterState":
        return InvalidTokenMinterState()
    if obj["kind"] == "ProgramPaused":
        return ProgramPaused()
    if obj["kind"] == "InvalidTokenPairState":
        return InvalidTokenPairState()
    if obj["kind"] == "InvalidLocalTokenState":
        return InvalidLocalTokenState()
    if obj["kind"] == "InvalidPauser":
        return InvalidPauser()
    if obj["kind"] == "InvalidTokenController":
        return InvalidTokenController()
    if obj["kind"] == "BurnAmountExceeded":
        return BurnAmountExceeded()
    if obj["kind"] == "InvalidAmount":
        return InvalidAmount()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "InvalidAuthority" / borsh.CStruct(),
    "InvalidTokenMinterState" / borsh.CStruct(),
    "ProgramPaused" / borsh.CStruct(),
    "InvalidTokenPairState" / borsh.CStruct(),
    "InvalidLocalTokenState" / borsh.CStruct(),
    "InvalidPauser" / borsh.CStruct(),
    "InvalidTokenController" / borsh.CStruct(),
    "BurnAmountExceeded" / borsh.CStruct(),
    "InvalidAmount" / borsh.CStruct(),
)



from .register_tristero_oapp import (
    register_tristero_oapp,
    RegisterTristeroOappArgs,
    RegisterTristeroOappAccounts,
)
from .admin_panel_update import (
    admin_panel_update,
    AdminPanelUpdateArgs,
    AdminPanelUpdateAccounts,
)
from .create_match import create_match, CreateMatchArgs, CreateMatchAccounts
from .execute_match import execute_match, ExecuteMatchArgs, ExecuteMatchAccounts
from .confirm_match import confirm_match, ConfirmMatchArgs, ConfirmMatchAccounts
from .start_challenge import start_challenge, StartChallengeArgs, StartChallengeAccounts
from .finish_challenge import (
    finish_challenge,
    FinishChallengeArgs,
    FinishChallengeAccounts,
)
from .lz_receive import lz_receive, LzReceiveArgs, LzReceiveAccounts
from .fill_swap import fill_swap, FillSwapArgs, FillSwapAccounts
from .lz_receive_types import (
    lz_receive_types,
    LzReceiveTypesArgs,
    LzReceiveTypesAccounts,
)
from .register_config import register_config, RegisterConfigArgs, RegisterConfigAccounts
from .place_order import place_order, PlaceOrderArgs, PlaceOrderAccounts
from .cancel_order import cancel_order, CancelOrderAccounts
from .unwind_match import unwind_match, UnwindMatchAccounts



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class ExecuteMatchParamsJSON(typing.TypedDict):
    dst_eid: int
    trade_match_id: int
    source_sell_amount: int
    sender: list[int]


@dataclass
class ExecuteMatchParams:
    layout: typing.ClassVar = borsh.CStruct(
        "dst_eid" / borsh.U32,
        "trade_match_id" / borsh.U64,
        "source_sell_amount" / borsh.U64,
        "sender" / borsh.U8[20],
    )
    dst_eid: int
    trade_match_id: int
    source_sell_amount: int
    sender: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "ExecuteMatchParams":
        return cls(
            dst_eid=obj.dst_eid,
            trade_match_id=obj.trade_match_id,
            source_sell_amount=obj.source_sell_amount,
            sender=obj.sender,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "dst_eid": self.dst_eid,
            "trade_match_id": self.trade_match_id,
            "source_sell_amount": self.source_sell_amount,
            "sender": self.sender,
        }

    def to_json(self) -> ExecuteMatchParamsJSON:
        return {
            "dst_eid": self.dst_eid,
            "trade_match_id": self.trade_match_id,
            "source_sell_amount": self.source_sell_amount,
            "sender": self.sender,
        }

    @classmethod
    def from_json(cls, obj: ExecuteMatchParamsJSON) -> "ExecuteMatchParams":
        return cls(
            dst_eid=obj["dst_eid"],
            trade_match_id=obj["trade_match_id"],
            source_sell_amount=obj["source_sell_amount"],
            sender=obj["sender"],
        )



from __future__ import annotations
import typing

from eth_typing import ChecksumAddress

from ...account import (
    Account,
    EthereumAccount,
    SolanaAccount,
    TronAccount,
)
from ...asset import SolanaToken, Token
from ...chain import Chain
from ...chain_client import (
    ChainClient,
    EthereumClient,
    SolanaClient,
    TronClient,
)
from ...transaction import Transaction
from ..types import OrderData
from . import ethereum, solana, tron


async def create_place_order_transaction[SrcChainType: Chain, DestChainType: Chain](
    *,
    src_client: ChainClient[SrcChainType],
    src_token: Token[SrcChainType],
    order_data: OrderData,
    account: Account[SrcChainType],
) -> Transaction[SrcChainType]:
    order_book_address = order_data.contract_address

    match src_client:
        case EthereumClient():
            return await ethereum.create_place_order_transaction(
                src_client=src_client,
                order_book_address=typing.cast(ChecksumAddress, order_book_address),
                order_data=order_data,
                account=typing.cast(EthereumAccount, account),
            )

        case SolanaClient():
            return await solana.create_place_order_transaction(
                src_client=src_client,
                src_token=typing.cast(SolanaToken, src_token),
                order_book_address=order_book_address,
                order_data=order_data,
                account=typing.cast(SolanaAccount, account),
            )

        case TronClient():
            return await tron.create_place_order_transaction(
                src_client=src_client,
                order_book_address=order_book_address,
                order_data=order_data,
                account=typing.cast(TronAccount, account),
            )

        case _:
            raise NotImplementedError(src_client.chain)


__all__ = ["create_place_order_transaction"]



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class LinkTokenPairParamsJSON(typing.TypedDict):
    local_token: str
    remote_domain: int
    remote_token: str


@dataclass
class LinkTokenPairParams:
    layout: typing.ClassVar = borsh.CStruct(
        "local_token" / BorshPubkey,
        "remote_domain" / borsh.U32,
        "remote_token" / BorshPubkey,
    )
    local_token: Pubkey
    remote_domain: int
    remote_token: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "LinkTokenPairParams":
        return cls(
            local_token=obj.local_token,
            remote_domain=obj.remote_domain,
            remote_token=obj.remote_token,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "local_token": self.local_token,
            "remote_domain": self.remote_domain,
            "remote_token": self.remote_token,
        }

    def to_json(self) -> LinkTokenPairParamsJSON:
        return {
            "local_token": str(self.local_token),
            "remote_domain": self.remote_domain,
            "remote_token": str(self.remote_token),
        }

    @classmethod
    def from_json(cls, obj: LinkTokenPairParamsJSON) -> "LinkTokenPairParams":
        return cls(
            local_token=Pubkey.from_string(obj["local_token"]),
            remote_domain=obj["remote_domain"],
            remote_token=Pubkey.from_string(obj["remote_token"]),
        )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class LzReceiveArgs(typing.TypedDict):
    params: types.lz_receive_params.LzReceiveParams


layout = borsh.CStruct("params" / types.lz_receive_params.LzReceiveParams.layout)


class LzReceiveAccounts(typing.TypedDict):
    oapp: Pubkey


def lz_receive(
    args: LzReceiveArgs,
    accounts: LzReceiveAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["oapp"], is_signer=False, is_writable=True)
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x08\xb3xm!v\xbdP"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class ExecuteMatchArgs(typing.TypedDict):
    params: types.execute_match_params.ExecuteMatchParams


layout = borsh.CStruct("params" / types.execute_match_params.ExecuteMatchParams.layout)


class ExecuteMatchAccounts(typing.TypedDict):
    authority: Pubkey
    token_account: Pubkey
    arb_user_token_account: Pubkey
    receipt: Pubkey


def execute_match(
    args: ExecuteMatchArgs,
    accounts: ExecuteMatchAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["token_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["arb_user_token_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["receipt"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"L/[\xdf\x14\n\x93\xe8"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from .admin_panel import AdminPanel, AdminPanelJSON
from .lz_receive_types_accounts import (
    LzReceiveTypesAccounts,
    LzReceiveTypesAccountsJSON,
)
from .order import Order, OrderJSON
from .receipt import Receipt, ReceiptJSON
from .trade_match import TradeMatch, TradeMatchJSON



from __future__ import annotations
import dataclasses
import typing

from solders.keypair import Keypair
from solders.pubkey import Pubkey

from ..chain import SolanaChain
from ..utility import solana
from .account import Account, AccountID


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class SolanaAccountID[Chain: SolanaChain](AccountID[Chain]):
    pubkey: Pubkey

    @classmethod
    def from_pubkey(cls, chain: Chain, pubkey: str) -> SolanaAccountID[Chain]:
        return cls(chain=chain, pubkey=Pubkey.from_string(pubkey))

    @property
    @typing.override
    def address(self) -> str:
        return str(self.pubkey)

    @typing.override
    def encode_address(self) -> bytes:
        return solana.encode_address(self.pubkey)


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class SolanaAccount[Chain: SolanaChain](Account[Chain]):
    keypair: Keypair

    @classmethod
    def from_private_key(cls, chain: Chain, private_key: str) -> SolanaAccount[Chain]:
        return cls(chain=chain, keypair=Keypair.from_base58_string(private_key))

    @property
    @typing.override
    def address(self) -> str:
        return str(self.keypair.pubkey())

    @property
    @typing.override
    def private_key(self) -> str:
        return str(self.keypair)

    @typing.override
    def encode_address(self) -> bytes:
        return solana.encode_address(self.keypair.pubkey())

    @typing.override
    def downcast(self) -> SolanaAccountID:
        return SolanaAccountID(chain=self.chain, pubkey=self.keypair.pubkey())



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class ReplaceDepositForBurnParamsJSON(typing.TypedDict):
    original_message: list[int]
    original_attestation: list[int]
    new_destination_caller: str
    new_mint_recipient: str


@dataclass
class ReplaceDepositForBurnParams:
    layout: typing.ClassVar = borsh.CStruct(
        "original_message" / borsh.Bytes,
        "original_attestation" / borsh.Bytes,
        "new_destination_caller" / BorshPubkey,
        "new_mint_recipient" / BorshPubkey,
    )
    original_message: bytes
    original_attestation: bytes
    new_destination_caller: Pubkey
    new_mint_recipient: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "ReplaceDepositForBurnParams":
        return cls(
            original_message=obj.original_message,
            original_attestation=obj.original_attestation,
            new_destination_caller=obj.new_destination_caller,
            new_mint_recipient=obj.new_mint_recipient,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "original_message": self.original_message,
            "original_attestation": self.original_attestation,
            "new_destination_caller": self.new_destination_caller,
            "new_mint_recipient": self.new_mint_recipient,
        }

    def to_json(self) -> ReplaceDepositForBurnParamsJSON:
        return {
            "original_message": list(self.original_message),
            "original_attestation": list(self.original_attestation),
            "new_destination_caller": str(self.new_destination_caller),
            "new_mint_recipient": str(self.new_mint_recipient),
        }

    @classmethod
    def from_json(
        cls, obj: ReplaceDepositForBurnParamsJSON
    ) -> "ReplaceDepositForBurnParams":
        return cls(
            original_message=bytes(obj["original_message"]),
            original_attestation=bytes(obj["original_attestation"]),
            new_destination_caller=Pubkey.from_string(obj["new_destination_caller"]),
            new_mint_recipient=Pubkey.from_string(obj["new_mint_recipient"]),
        )



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class RemoteTokenMessengerJSON(typing.TypedDict):
    domain: int
    token_messenger: str


@dataclass
class RemoteTokenMessenger:
    discriminator: typing.ClassVar = b'is\xae"_\xe9\x8a\xfc'
    layout: typing.ClassVar = borsh.CStruct(
        "domain" / borsh.U32, "token_messenger" / BorshPubkey
    )
    domain: int
    token_messenger: Pubkey

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["RemoteTokenMessenger"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["RemoteTokenMessenger"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["RemoteTokenMessenger"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "RemoteTokenMessenger":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = RemoteTokenMessenger.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            domain=dec.domain,
            token_messenger=dec.token_messenger,
        )

    def to_json(self) -> RemoteTokenMessengerJSON:
        return {
            "domain": self.domain,
            "token_messenger": str(self.token_messenger),
        }

    @classmethod
    def from_json(cls, obj: RemoteTokenMessengerJSON) -> "RemoteTokenMessenger":
        return cls(
            domain=obj["domain"],
            token_messenger=Pubkey.from_string(obj["token_messenger"]),
        )



from .account import Account, AccountBase, AccountID
from .account_manager import (
    AccountManager,
    AccountIDManager,
    HDWalletAccountManager,
    SimpleAccountIDManager,
    SimpleAccountManager,
)
from .ethereum import EthereumAccount, EthereumAccountID
from .hdwallet import Wallet
from .solana import SolanaAccount, SolanaAccountID
from .tron import TronAccount, TronAccountID


__all__ = [
    "Account",
    "AccountBase",
    "AccountID",
    "AccountIDManager",
    "AccountManager",
    "EthereumAccount",
    "EthereumAccountID",
    "HDWalletAccountManager",
    "SimpleAccountIDManager",
    "SimpleAccountManager",
    "SolanaAccount",
    "SolanaAccountID",
    "TronAccount",
    "TronAccountID",
    "Wallet",
]



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from ...chain_client import ChainClient
from ...transaction import SentTransaction
from .types import Chain


class CCTPRequest(BaseModel):
    chain: Chain
    burn_tx: str


class GasResponse(BaseModel):
    gas_estimate: int
    gas_price: int


class OrderRequest(BaseModel):
    chain: Chain
    place_taker_tx: str
    referral_code: Optional[str] = None


class OrderType(str, Enum):
    DEFAULT = "default"
    CCTP = "cctp"


class OrderResponse(BaseModel):
    id: str
    taker_address: str
    maker_address: Optional[str] = None
    src_asset_address: str
    dst_asset_address: str
    src_chain: Chain
    dst_chain: Chain
    src_amount: int
    dst_amount: int
    created_at: str
    filled_at: Optional[str] = None
    expires_at: Optional[str] = None
    completed: bool
    place_tx: str
    fill_tx: Optional[str] = None
    eta: int = 0
    order_type: OrderType = OrderType.DEFAULT

    async def to_sent_transaction(self) -> SentTransaction:
        client = await ChainClient.create(self.src_chain.to_chain())
        return SentTransaction.from_id(client, self.place_tx)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class FillSwapParamsJSON(typing.TypedDict):
    order_idx: int
    trade_match_id: int
    src_quantity: int
    dst_quantity: int


@dataclass
class FillSwapParams:
    layout: typing.ClassVar = borsh.CStruct(
        "order_idx" / borsh.U64,
        "trade_match_id" / borsh.U64,
        "src_quantity" / borsh.U64,
        "dst_quantity" / borsh.U64,
    )
    order_idx: int
    trade_match_id: int
    src_quantity: int
    dst_quantity: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "FillSwapParams":
        return cls(
            order_idx=obj.order_idx,
            trade_match_id=obj.trade_match_id,
            src_quantity=obj.src_quantity,
            dst_quantity=obj.dst_quantity,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "order_idx": self.order_idx,
            "trade_match_id": self.trade_match_id,
            "src_quantity": self.src_quantity,
            "dst_quantity": self.dst_quantity,
        }

    def to_json(self) -> FillSwapParamsJSON:
        return {
            "order_idx": self.order_idx,
            "trade_match_id": self.trade_match_id,
            "src_quantity": self.src_quantity,
            "dst_quantity": self.dst_quantity,
        }

    @classmethod
    def from_json(cls, obj: FillSwapParamsJSON) -> "FillSwapParams":
        return cls(
            order_idx=obj["order_idx"],
            trade_match_id=obj["trade_match_id"],
            src_quantity=obj["src_quantity"],
            dst_quantity=obj["dst_quantity"],
        )



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class TokenPairJSON(typing.TypedDict):
    remote_domain: int
    remote_token: str
    local_token: str
    bump: int


@dataclass
class TokenPair:
    discriminator: typing.ClassVar = b"\x11\xd6-\xb0\xe5\x95\xc5G"
    layout: typing.ClassVar = borsh.CStruct(
        "remote_domain" / borsh.U32,
        "remote_token" / BorshPubkey,
        "local_token" / BorshPubkey,
        "bump" / borsh.U8,
    )
    remote_domain: int
    remote_token: Pubkey
    local_token: Pubkey
    bump: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["TokenPair"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["TokenPair"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["TokenPair"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "TokenPair":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = TokenPair.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            remote_domain=dec.remote_domain,
            remote_token=dec.remote_token,
            local_token=dec.local_token,
            bump=dec.bump,
        )

    def to_json(self) -> TokenPairJSON:
        return {
            "remote_domain": self.remote_domain,
            "remote_token": str(self.remote_token),
            "local_token": str(self.local_token),
            "bump": self.bump,
        }

    @classmethod
    def from_json(cls, obj: TokenPairJSON) -> "TokenPair":
        return cls(
            remote_domain=obj["remote_domain"],
            remote_token=Pubkey.from_string(obj["remote_token"]),
            local_token=Pubkey.from_string(obj["local_token"]),
            bump=obj["bump"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class InitializeParamsJSON(typing.TypedDict):
    token_controller: str
    local_message_transmitter: str
    message_body_version: int


@dataclass
class InitializeParams:
    layout: typing.ClassVar = borsh.CStruct(
        "token_controller" / BorshPubkey,
        "local_message_transmitter" / BorshPubkey,
        "message_body_version" / borsh.U32,
    )
    token_controller: Pubkey
    local_message_transmitter: Pubkey
    message_body_version: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeParams":
        return cls(
            token_controller=obj.token_controller,
            local_message_transmitter=obj.local_message_transmitter,
            message_body_version=obj.message_body_version,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "token_controller": self.token_controller,
            "local_message_transmitter": self.local_message_transmitter,
            "message_body_version": self.message_body_version,
        }

    def to_json(self) -> InitializeParamsJSON:
        return {
            "token_controller": str(self.token_controller),
            "local_message_transmitter": str(self.local_message_transmitter),
            "message_body_version": self.message_body_version,
        }

    @classmethod
    def from_json(cls, obj: InitializeParamsJSON) -> "InitializeParams":
        return cls(
            token_controller=Pubkey.from_string(obj["token_controller"]),
            local_message_transmitter=Pubkey.from_string(
                obj["local_message_transmitter"]
            ),
            message_body_version=obj["message_body_version"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class FinishChallengeParamsJSON(typing.TypedDict):
    arb_eid: int
    trade_match_id: int
    spl_token: str
    erc20token: list[int]
    receiver: list[int]


@dataclass
class FinishChallengeParams:
    layout: typing.ClassVar = borsh.CStruct(
        "arb_eid" / borsh.U32,
        "trade_match_id" / borsh.U64,
        "spl_token" / BorshPubkey,
        "erc20token" / borsh.U8[20],
        "receiver" / borsh.U8[32],
    )
    arb_eid: int
    trade_match_id: int
    spl_token: Pubkey
    erc20token: list[int]
    receiver: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "FinishChallengeParams":
        return cls(
            arb_eid=obj.arb_eid,
            trade_match_id=obj.trade_match_id,
            spl_token=obj.spl_token,
            erc20token=obj.erc20token,
            receiver=obj.receiver,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "arb_eid": self.arb_eid,
            "trade_match_id": self.trade_match_id,
            "spl_token": self.spl_token,
            "erc20token": self.erc20token,
            "receiver": self.receiver,
        }

    def to_json(self) -> FinishChallengeParamsJSON:
        return {
            "arb_eid": self.arb_eid,
            "trade_match_id": self.trade_match_id,
            "spl_token": str(self.spl_token),
            "erc20token": self.erc20token,
            "receiver": self.receiver,
        }

    @classmethod
    def from_json(cls, obj: FinishChallengeParamsJSON) -> "FinishChallengeParams":
        return cls(
            arb_eid=obj["arb_eid"],
            trade_match_id=obj["trade_match_id"],
            spl_token=Pubkey.from_string(obj["spl_token"]),
            erc20token=obj["erc20token"],
            receiver=obj["receiver"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class ChallengeParamsJSON(typing.TypedDict):
    trade_match_id: int
    tristero_oapp_bump: int
    receiver: list[int]
    taker: list[int]


@dataclass
class ChallengeParams:
    layout: typing.ClassVar = borsh.CStruct(
        "trade_match_id" / borsh.U64,
        "tristero_oapp_bump" / borsh.U8,
        "receiver" / borsh.U8[32],
        "taker" / borsh.U8[20],
    )
    trade_match_id: int
    tristero_oapp_bump: int
    receiver: list[int]
    taker: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "ChallengeParams":
        return cls(
            trade_match_id=obj.trade_match_id,
            tristero_oapp_bump=obj.tristero_oapp_bump,
            receiver=obj.receiver,
            taker=obj.taker,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "trade_match_id": self.trade_match_id,
            "tristero_oapp_bump": self.tristero_oapp_bump,
            "receiver": self.receiver,
            "taker": self.taker,
        }

    def to_json(self) -> ChallengeParamsJSON:
        return {
            "trade_match_id": self.trade_match_id,
            "tristero_oapp_bump": self.tristero_oapp_bump,
            "receiver": self.receiver,
            "taker": self.taker,
        }

    @classmethod
    def from_json(cls, obj: ChallengeParamsJSON) -> "ChallengeParams":
        return cls(
            trade_match_id=obj["trade_match_id"],
            tristero_oapp_bump=obj["tristero_oapp_bump"],
            receiver=obj["receiver"],
            taker=obj["taker"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class TransferOwnershipParamsJSON(typing.TypedDict):
    new_owner: str


@dataclass
class TransferOwnershipParams:
    layout: typing.ClassVar = borsh.CStruct("new_owner" / BorshPubkey)
    new_owner: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "TransferOwnershipParams":
        return cls(new_owner=obj.new_owner)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"new_owner": self.new_owner}

    def to_json(self) -> TransferOwnershipParamsJSON:
        return {"new_owner": str(self.new_owner)}

    @classmethod
    def from_json(cls, obj: TransferOwnershipParamsJSON) -> "TransferOwnershipParams":
        return cls(new_owner=Pubkey.from_string(obj["new_owner"]))



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class LocalTokenJSON(typing.TypedDict):
    custody: str
    mint: str
    burn_limit_per_message: int
    messages_sent: int
    messages_received: int
    amount_sent: int
    amount_received: int
    bump: int
    custody_bump: int


@dataclass
class LocalToken:
    discriminator: typing.ClassVar = b"\x9f\x83:\xaa\xc1T\x80\xb6"
    layout: typing.ClassVar = borsh.CStruct(
        "custody" / BorshPubkey,
        "mint" / BorshPubkey,
        "burn_limit_per_message" / borsh.U64,
        "messages_sent" / borsh.U64,
        "messages_received" / borsh.U64,
        "amount_sent" / borsh.U128,
        "amount_received" / borsh.U128,
        "bump" / borsh.U8,
        "custody_bump" / borsh.U8,
    )
    custody: Pubkey
    mint: Pubkey
    burn_limit_per_message: int
    messages_sent: int
    messages_received: int
    amount_sent: int
    amount_received: int
    bump: int
    custody_bump: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["LocalToken"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["LocalToken"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["LocalToken"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "LocalToken":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = LocalToken.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            custody=dec.custody,
            mint=dec.mint,
            burn_limit_per_message=dec.burn_limit_per_message,
            messages_sent=dec.messages_sent,
            messages_received=dec.messages_received,
            amount_sent=dec.amount_sent,
            amount_received=dec.amount_received,
            bump=dec.bump,
            custody_bump=dec.custody_bump,
        )

    def to_json(self) -> LocalTokenJSON:
        return {
            "custody": str(self.custody),
            "mint": str(self.mint),
            "burn_limit_per_message": self.burn_limit_per_message,
            "messages_sent": self.messages_sent,
            "messages_received": self.messages_received,
            "amount_sent": self.amount_sent,
            "amount_received": self.amount_received,
            "bump": self.bump,
            "custody_bump": self.custody_bump,
        }

    @classmethod
    def from_json(cls, obj: LocalTokenJSON) -> "LocalToken":
        return cls(
            custody=Pubkey.from_string(obj["custody"]),
            mint=Pubkey.from_string(obj["mint"]),
            burn_limit_per_message=obj["burn_limit_per_message"],
            messages_sent=obj["messages_sent"],
            messages_received=obj["messages_received"],
            amount_sent=obj["amount_sent"],
            amount_received=obj["amount_received"],
            bump=obj["bump"],
            custody_bump=obj["custody_bump"],
        )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class FinishChallengeArgs(typing.TypedDict):
    params: types.finish_challenge_params.FinishChallengeParams


layout = borsh.CStruct(
    "params" / types.finish_challenge_params.FinishChallengeParams.layout
)


class FinishChallengeAccounts(typing.TypedDict):
    authority: Pubkey
    oapp: Pubkey


def finish_challenge(
    args: FinishChallengeArgs,
    accounts: FinishChallengeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["oapp"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xbap\xaa\xc4\xb4{Ip"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class ConfirmMatchParamsJSON(typing.TypedDict):
    trade_match_id: int


@dataclass
class ConfirmMatchParams:
    layout: typing.ClassVar = borsh.CStruct("trade_match_id" / borsh.U64)
    trade_match_id: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "ConfirmMatchParams":
        return cls(trade_match_id=obj.trade_match_id)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"trade_match_id": self.trade_match_id}

    def to_json(self) -> ConfirmMatchParamsJSON:
        return {"trade_match_id": self.trade_match_id}

    @classmethod
    def from_json(cls, obj: ConfirmMatchParamsJSON) -> "ConfirmMatchParams":
        return cls(trade_match_id=obj["trade_match_id"])



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class RegisterConfigArgs(typing.TypedDict):
    param_pubkey: Pubkey


layout = borsh.CStruct("param_pubkey" / BorshPubkey)


class RegisterConfigAccounts(typing.TypedDict):
    payer: Pubkey
    oapp_config: Pubkey
    lz_receive_types_accounts: Pubkey


def register_config(
    args: RegisterConfigArgs,
    accounts: RegisterConfigAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["oapp_config"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["lz_receive_types_accounts"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b" \xf7R\x83#\xb7\x079"
    encoded_args = layout.build(
        {
            "param_pubkey": args["param_pubkey"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



# Reference:
# https://github.com/tristeroresearch/cache-half-full/blob/main/backend/book/svm_book.py
# https://github.com/tristeroresearch/dapp/blob/dev/src/chains/solana/placeOrder.ts

from anchorpy.program.core import Program
from anchorpy.provider import Provider, Wallet
from solders.pubkey import Pubkey

from .... import config
from ....account import SolanaAccount
from ....asset import SolanaToken
from ....chain import SolanaChain
from ....chain_client import SolanaClient
from ....utility import solana
from ....transaction import SolanaTransaction
from ...types import OrderData
from .anchor import instructions
from .anchor.types import PlaceOrderParams
from .anchor.instructions import PlaceOrderAccounts
from .anchor.instructions.place_order import PlaceOrderArgs


def to_bytes_list(string: str) -> list[int]:
    return list(bytes(Pubkey.from_string(string)).rjust(32, b"\0"))


def make_order_book_program[ChainType: SolanaChain](
    program_id: Pubkey,
    client: SolanaClient[ChainType],
    account: SolanaAccount[ChainType],
) -> Program:
    wallet = Wallet(payer=account.keypair)
    provider = Provider(connection=client.native, wallet=wallet)
    return Program(idl=config.tristero_idl, program_id=program_id, provider=provider)


async def make_order_pda(program: Program, admin_panel: Pubkey) -> Pubkey:
    admin_panel_account = await program.account["AdminPanel"].fetch(admin_panel)
    order_count = admin_panel_account.order_count
    order_id_buffer = order_count.to_bytes(8, byteorder="big")
    return solana.find_program_address("order", program.program_id, (order_id_buffer,))


async def create_place_order_transaction[
    SrcChainType: SolanaChain,
](
    *,
    src_client: SolanaClient[SrcChainType],
    src_token: SolanaToken[SrcChainType],
    order_book_address: str,
    order_data: OrderData,
    account: SolanaAccount[SrcChainType],
) -> SolanaTransaction[SrcChainType]:
    program = make_order_book_program(
        program_id=Pubkey.from_string(order_book_address),
        client=src_client,
        account=account,
    )

    wallet_pubkey = account.keypair.pubkey()

    oapp = solana.find_program_address("TristeroOapp", program.program_id)

    admin_panel = solana.find_program_address("admin_panel", program.program_id)

    token_account = src_token.associated_token_account(account.keypair.pubkey())

    order_pda = await make_order_pda(program, admin_panel)

    accounts = PlaceOrderAccounts(
        authority=wallet_pubkey,
        oapp=oapp,
        admin_panel=admin_panel,
        token_mint=src_token.mint,
        token_account=token_account,
        match_account=wallet_pubkey,
        order=order_pda,
    )

    args = PlaceOrderArgs(
        params=PlaceOrderParams(
            source_sell_amount=order_data.order_funding.src_amount_in,
            min_sell_amount=order_data.order_funding.src_amount_in,
            dst_token_mint=to_bytes_list(order_data.order_direction.dst_token_address),
            dst_buy_amount=order_data.order_funding.dst_amount_out,
            eid=order_data.order_direction.dst_lzc,
            target_address=to_bytes_list(order_data.target_address),
            bond_asset_mint=Pubkey.from_string(
                order_data.order_funding.bond_token_address
            ),
            bond_amount=order_data.order_funding.bond_amount,
            bond_fee=order_data.order_funding.bond_fee,
        )
    )

    instruction = instructions.place_order(args=args, accounts=accounts)

    return await SolanaTransaction.create(
        src_client,
        (instruction,),
        (account,),
    )



import typing

from ..asset import Token
from ..chain import Chain, SupportedChain
from ..transaction import SentTransaction
from .scanner import Scanner


SCANNER_URLS: dict[Chain, str] = {
    SupportedChain.ETHEREUM.value: "https://etherscan.io",
    SupportedChain.OPTIMISM.value: "https://optimistic.etherscan.io",
    SupportedChain.BSC.value: "https://bscscan.com",
    SupportedChain.POLYGON.value: "https://polygonscan.com",
    SupportedChain.OPBNB.value: "https://opbnbscan.com",
    SupportedChain.MANTLE.value: "https://explorer.mantle.xyz",
    SupportedChain.BASE.value: "https://basescan.org",
    SupportedChain.MODE.value: "https://modescan.io",
    SupportedChain.ARBITRUM.value: "https://arbiscan.io",
    SupportedChain.CELO.value: "https://explorer.celo.org/mainnet",
    SupportedChain.AVALANCHE_C_CHAIN.value: "https://snowscan.xyz",
    SupportedChain.BLAST.value: "https://blastscan.io",
    SupportedChain.SCROLL.value: "https://scrollscan.com",
}


# EIP-3091 Blockchain Scanner
# https://eips.ethereum.org/EIPS/eip-3091
class EIP3091Scanner(Scanner):
    __slots__ = ("url",)

    def __init__(self, chain: Chain) -> None:
        self.url = SCANNER_URLS[chain]

    @property
    @typing.override
    def base_url(self) -> str:
        return self.url

    @typing.override
    def address(self, address: str) -> str:
        return f"{self.url}/address/{address}"

    @typing.override
    def transaction(self, transaction: SentTransaction) -> str:
        return f"{self.url}/tx/{transaction.id}"

    @typing.override
    def token(self, token: Token) -> str:
        return f"{self.url}/token/{token.address}"



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class UpdateParamsJSON(typing.TypedDict):
    admin_wallet: str
    payment_wallet: str


@dataclass
class UpdateParams:
    layout: typing.ClassVar = borsh.CStruct(
        "admin_wallet" / BorshPubkey, "payment_wallet" / BorshPubkey
    )
    admin_wallet: Pubkey
    payment_wallet: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateParams":
        return cls(admin_wallet=obj.admin_wallet, payment_wallet=obj.payment_wallet)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "admin_wallet": self.admin_wallet,
            "payment_wallet": self.payment_wallet,
        }

    def to_json(self) -> UpdateParamsJSON:
        return {
            "admin_wallet": str(self.admin_wallet),
            "payment_wallet": str(self.payment_wallet),
        }

    @classmethod
    def from_json(cls, obj: UpdateParamsJSON) -> "UpdateParams":
        return cls(
            admin_wallet=Pubkey.from_string(obj["admin_wallet"]),
            payment_wallet=Pubkey.from_string(obj["payment_wallet"]),
        )



from pydantic import BaseModel


class OrderDirection(BaseModel):
    src_token_address: str
    dst_token_address: str
    src_lzc: int
    dst_lzc: int

    def to_eth(self) -> dict:
        return {
            "srcAsset": self.src_token_address,
            "dstAsset": self.dst_token_address,
            "dstLzc": self.dst_lzc,
        }

    def to_tron(self) -> list:
        return [
            self.src_token_address,
            self.dst_token_address,
            self.dst_lzc,
        ]


class OrderFunding(BaseModel):
    src_amount_in: int
    dst_amount_out: int
    bond_token_address: str
    bond_amount: int
    bond_fee: int

    def to_eth(self) -> dict:
        return {
            "srcQuantity": self.src_amount_in,
            "dstQuantity": self.dst_amount_out,
            "bondFee": self.bond_fee,
            "bondAsset": self.bond_token_address,
            "bondAmount": self.bond_amount,
        }

    def to_tron(self) -> list:
        return [
            self.src_amount_in,
            self.dst_amount_out,
            self.bond_fee,
            self.bond_token_address,
            self.bond_amount,
        ]


class OrderExpiration(BaseModel):
    timestamp: int
    challenge_offset: int
    challenge_window: int

    def to_eth(self) -> dict:
        return {
            "timestamp": self.timestamp,
            "challengeOffset": self.challenge_offset,
            "challengeWindow": self.challenge_window,
        }

    def to_tron(self) -> list:
        return [
            self.timestamp,
            self.challenge_offset,
            self.challenge_window,
        ]



from __future__ import annotations
import asyncio
import dataclasses
import pprint
import typing
from typing import ClassVar

from eth_account.datastructures import SignedTransaction
from hexbytes import HexBytes
from web3 import AsyncWeb3
from web3._utils import async_transactions
from web3.contract.async_contract import AsyncContractFunction
from web3.types import TxParams, TxReceipt

from ..account import EthereumAccount
from ..chain import EthereumChain
from ..chain_client import EthereumClient
from .transaction import SentTransaction, Transaction


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class EthereumSentTransaction[Chain: EthereumChain](SentTransaction[Chain]):
    client: EthereumClient[Chain]
    transaction_hash: HexBytes

    @property
    @typing.override
    def id(self) -> str:
        return self.transaction_hash.to_0x_hex()

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @typing.override
    async def wait_for_receipt(self, **kwargs) -> TxReceipt:
        receipt = await self.client.w3.eth.wait_for_transaction_receipt(
            self.transaction_hash, **kwargs
        )
        assert (
            receipt["status"] == 0x1
        ), f"Transaction failed:\n{pprint.pformat(receipt)}"
        return receipt


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class EthereumTransaction[Chain: EthereumChain](Transaction[Chain]):
    DEFAULT_GAS_MULTIPLIER: ClassVar[float] = 1.5

    client: EthereumClient[Chain]
    transaction: SignedTransaction

    @staticmethod
    async def fill_transaction_defaults(
        w3: AsyncWeb3,
        params: TxParams = TxParams(),
    ) -> TxParams:
        return await async_transactions.async_fill_transaction_defaults(w3, params)

    @classmethod
    async def from_transaction_params(
        cls,
        client: EthereumClient[Chain],
        params: TxParams,
        signer: EthereumAccount[Chain],
        gas_multiplier: float = DEFAULT_GAS_MULTIPLIER,
    ) -> EthereumTransaction[Chain]:
        params, nonce = await asyncio.gather(
            async_transactions.async_fill_transaction_defaults(client.w3, params),
            client.w3.eth.get_transaction_count(signer.native.address, "latest"),
        )

        params.update(
            # Ensures the transaction goes through
            gas=int(gas_multiplier * params["gas"]),  # type: ignore
            # This isn't filled in by `async_fill_transaction_defaults`
            nonce=nonce,
        )

        signed_transaction = signer.native.sign_transaction(params)  # type: ignore

        return cls(client=client, transaction=signed_transaction)

    @classmethod
    async def from_contract_function(
        cls,
        client: EthereumClient[Chain],
        contract_function: AsyncContractFunction,
        signer: EthereumAccount,
        gas_multiplier: float = DEFAULT_GAS_MULTIPLIER,
    ) -> EthereumTransaction[Chain]:
        params, nonce = await asyncio.gather(
            contract_function.build_transaction({"from": signer.native.address}),
            client.w3.eth.get_transaction_count(signer.native.address, "latest"),
        )

        params.update(
            # Ensures the transaction goes through
            gas=int(gas_multiplier * params["gas"]),  # type: ignore
            # This isn't filled in by `build_transaction`
            nonce=nonce,
        )

        signed_transaction = signer.native.sign_transaction(params)  # type: ignore

        return cls(client=client, transaction=signed_transaction)

    @typing.override
    async def broadcast(self) -> EthereumSentTransaction[Chain]:
        transaction_hash = await self.client.w3.eth.send_raw_transaction(
            self.transaction.raw_transaction
        )
        return EthereumSentTransaction(
            client=self.client, transaction_hash=transaction_hash
        )

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain



from .initialize import initialize, InitializeArgs, InitializeAccounts
from .transfer_ownership import (
    transfer_ownership,
    TransferOwnershipArgs,
    TransferOwnershipAccounts,
)
from .accept_ownership import (
    accept_ownership,
    AcceptOwnershipArgs,
    AcceptOwnershipAccounts,
)
from .add_remote_token_messenger import (
    add_remote_token_messenger,
    AddRemoteTokenMessengerArgs,
    AddRemoteTokenMessengerAccounts,
)
from .remove_remote_token_messenger import (
    remove_remote_token_messenger,
    RemoveRemoteTokenMessengerArgs,
    RemoveRemoteTokenMessengerAccounts,
)
from .deposit_for_burn import (
    deposit_for_burn,
    DepositForBurnArgs,
    DepositForBurnAccounts,
)
from .deposit_for_burn_with_caller import (
    deposit_for_burn_with_caller,
    DepositForBurnWithCallerArgs,
    DepositForBurnWithCallerAccounts,
)
from .replace_deposit_for_burn import (
    replace_deposit_for_burn,
    ReplaceDepositForBurnArgs,
    ReplaceDepositForBurnAccounts,
)
from .handle_receive_message import (
    handle_receive_message,
    HandleReceiveMessageArgs,
    HandleReceiveMessageAccounts,
)
from .set_token_controller import (
    set_token_controller,
    SetTokenControllerArgs,
    SetTokenControllerAccounts,
)
from .pause import pause, PauseArgs, PauseAccounts
from .unpause import unpause, UnpauseArgs, UnpauseAccounts
from .update_pauser import update_pauser, UpdatePauserArgs, UpdatePauserAccounts
from .set_max_burn_amount_per_message import (
    set_max_burn_amount_per_message,
    SetMaxBurnAmountPerMessageArgs,
    SetMaxBurnAmountPerMessageAccounts,
)
from .add_local_token import add_local_token, AddLocalTokenArgs, AddLocalTokenAccounts
from .remove_local_token import (
    remove_local_token,
    RemoveLocalTokenArgs,
    RemoveLocalTokenAccounts,
)
from .link_token_pair import link_token_pair, LinkTokenPairArgs, LinkTokenPairAccounts
from .unlink_token_pair import (
    unlink_token_pair,
    UnlinkTokenPairArgs,
    UnlinkTokenPairAccounts,
)
from .burn_token_custody import (
    burn_token_custody,
    BurnTokenCustodyArgs,
    BurnTokenCustodyAccounts,
)



import typing

from tronpy.async_contract import AsyncContractMethod

from ... import config
from ...account import TronAccount
from ...chain import TronChain
from ...chain_client import TronClient
from ...transaction import TronTransaction
from ..types import OrderData


async def create_place_order_transaction[SrcChainType: TronChain](
    *,
    src_client: TronClient[SrcChainType],
    order_book_address: str,
    order_data: OrderData,
    account: TronAccount[SrcChainType],
) -> TronTransaction[SrcChainType]:
    contract = await src_client.native.get_contract(order_book_address)
    contract.abi = config.tron_order_book_abi

    return await TronTransaction.from_contract_method(
        src_client,
        account,
        typing.cast(AsyncContractMethod, contract.functions.placeOrder),
        order_data.order_direction.to_tron(),
        order_data.order_funding.to_tron(),
        order_data.order_expiration.to_tron(),
        # TODO: bytes32
        order_data.target_address,
        order_data.filler_address,
    )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class FillSwapArgs(typing.TypedDict):
    params: types.fill_swap_params.FillSwapParams


layout = borsh.CStruct("params" / types.fill_swap_params.FillSwapParams.layout)


class FillSwapAccounts(typing.TypedDict):
    authority: Pubkey
    oapp: Pubkey
    order: Pubkey
    trade_match: Pubkey
    staking_account: Pubkey
    staking_bond_account: Pubkey
    user_token_account: Pubkey
    bonder_bond_token_account: Pubkey


def fill_swap(
    args: FillSwapArgs,
    accounts: FillSwapAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["oapp"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["order"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["trade_match"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["staking_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["staking_bond_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["user_token_account"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["bonder_bond_token_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xcc\xa5\xacV/^\xa8\x03"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class TradeMatchJSON(typing.TypedDict):
    authority: str
    user_token_addr: str
    source_token_mint: str
    dst_token_mint: list[int]
    arb_user_token_account: str
    order_idx: int
    source_sell_amount: int
    dst_buy_amount: int
    eid: int
    bump: int
    trade_match_id: int
    status: int


@dataclass
class TradeMatch:
    discriminator: typing.ClassVar = b"\x00\xa3a\t\x9a\xee\x01\xb1"
    layout: typing.ClassVar = borsh.CStruct(
        "authority" / BorshPubkey,
        "user_token_addr" / BorshPubkey,
        "source_token_mint" / BorshPubkey,
        "dst_token_mint" / borsh.U8[32],
        "arb_user_token_account" / BorshPubkey,
        "order_idx" / borsh.U64,
        "source_sell_amount" / borsh.U64,
        "dst_buy_amount" / borsh.U64,
        "eid" / borsh.U32,
        "bump" / borsh.U8,
        "trade_match_id" / borsh.U64,
        "status" / borsh.U8,
    )
    authority: Pubkey
    user_token_addr: Pubkey
    source_token_mint: Pubkey
    dst_token_mint: list[int]
    arb_user_token_account: Pubkey
    order_idx: int
    source_sell_amount: int
    dst_buy_amount: int
    eid: int
    bump: int
    trade_match_id: int
    status: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["TradeMatch"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["TradeMatch"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["TradeMatch"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "TradeMatch":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = TradeMatch.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            authority=dec.authority,
            user_token_addr=dec.user_token_addr,
            source_token_mint=dec.source_token_mint,
            dst_token_mint=dec.dst_token_mint,
            arb_user_token_account=dec.arb_user_token_account,
            order_idx=dec.order_idx,
            source_sell_amount=dec.source_sell_amount,
            dst_buy_amount=dec.dst_buy_amount,
            eid=dec.eid,
            bump=dec.bump,
            trade_match_id=dec.trade_match_id,
            status=dec.status,
        )

    def to_json(self) -> TradeMatchJSON:
        return {
            "authority": str(self.authority),
            "user_token_addr": str(self.user_token_addr),
            "source_token_mint": str(self.source_token_mint),
            "dst_token_mint": self.dst_token_mint,
            "arb_user_token_account": str(self.arb_user_token_account),
            "order_idx": self.order_idx,
            "source_sell_amount": self.source_sell_amount,
            "dst_buy_amount": self.dst_buy_amount,
            "eid": self.eid,
            "bump": self.bump,
            "trade_match_id": self.trade_match_id,
            "status": self.status,
        }

    @classmethod
    def from_json(cls, obj: TradeMatchJSON) -> "TradeMatch":
        return cls(
            authority=Pubkey.from_string(obj["authority"]),
            user_token_addr=Pubkey.from_string(obj["user_token_addr"]),
            source_token_mint=Pubkey.from_string(obj["source_token_mint"]),
            dst_token_mint=obj["dst_token_mint"],
            arb_user_token_account=Pubkey.from_string(obj["arb_user_token_account"]),
            order_idx=obj["order_idx"],
            source_sell_amount=obj["source_sell_amount"],
            dst_buy_amount=obj["dst_buy_amount"],
            eid=obj["eid"],
            bump=obj["bump"],
            trade_match_id=obj["trade_match_id"],
            status=obj["status"],
        )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class LzReceiveTypesArgs(typing.TypedDict):
    params: types.lz_receive_type_params.LzReceiveTypeParams


layout = borsh.CStruct(
    "params" / types.lz_receive_type_params.LzReceiveTypeParams.layout
)


class LzReceiveTypesAccounts(typing.TypedDict):
    oft_config: Pubkey
    message_lib: Pubkey


def lz_receive_types(
    args: LzReceiveTypesArgs,
    accounts: LzReceiveTypesAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["oft_config"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["message_lib"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdd\x11\xf6\x9f\xf8\x80\x1f`"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import dataclasses
import typing

from web3 import AsyncWeb3
from web3.middleware import ExtraDataToPOAMiddleware
from web3.providers import (
    AsyncBaseProvider,
    AsyncHTTPProvider,
    AsyncIPCProvider,
    PersistentConnectionProvider,
    WebSocketProvider,
)


from .. import config
from ..chain import EthereumChain
from .chain_client import ChainClient


async def make_provider(endpoint_uri: str) -> AsyncBaseProvider:
    if endpoint_uri.startswith("ws://") or endpoint_uri.startswith("wss://"):
        provider = WebSocketProvider(endpoint_uri, {"timeout": 60.0})
    elif endpoint_uri.startswith("http://") or endpoint_uri.startswith("https://"):
        provider = AsyncHTTPProvider(endpoint_uri, timeout=60.0)
    elif endpoint_uri.endswith(".ipc"):
        provider = AsyncIPCProvider(endpoint_uri)
    else:
        raise ValueError(f"Invalid endpoint URI: {endpoint_uri}")

    if isinstance(provider, PersistentConnectionProvider):
        await provider.connect()

    return provider


async def make_w3(chain: EthereumChain) -> AsyncWeb3:
    provider = await make_provider(config.config.endpoint_uris[chain])
    w3 = AsyncWeb3(provider)
    w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
    return w3


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class EthereumClient[Chain: EthereumChain](ChainClient[Chain]):
    w3: AsyncWeb3

    @classmethod
    @typing.override
    async def create(cls, chain: Chain) -> EthereumClient[Chain]:
        w3 = await make_w3(chain)
        return cls(chain=chain, w3=w3)

    @typing.override
    async def close(self) -> None:
        if await self.is_connected():
            await self.w3.provider.disconnect()

    @typing.override
    async def is_connected(self) -> bool:
        return await self.w3.is_connected()

    @typing.override    
    async def reconnect(self) -> None:
        if not isinstance(self.w3.provider, PersistentConnectionProvider):
            return

        await self.w3.provider.connect()


import dataclasses
from datetime import timedelta
from typing import Any

from ..account import AccountID
from ..asset import Token
from ..transaction import SentTransaction, Transaction
from .risk_manager import RiskAnalysis
from .types import OrderResponse, Quote


@dataclasses.dataclass(frozen=True, slots=True)
class InsufficientSourceBalance:
    token: Token
    balance: int
    account_id: AccountID


@dataclasses.dataclass(frozen=True, slots=True)
class QuoteFailed:
    pair: tuple[Token, Token]
    amount: int
    account_id: AccountID
    exception: Exception


@dataclasses.dataclass(frozen=True, slots=True)
class QuoteLiquidityUnavailable:
    pair: tuple[Token, Token]
    amount: int
    account_id: AccountID
    quote: Quote


@dataclasses.dataclass(frozen=True, slots=True)
class OrderRejected:
    pair: tuple[Token, Token]
    amount: int
    account_id: AccountID
    quote: Quote


@dataclasses.dataclass(frozen=True, slots=True)
class RiskManagerRejection:
    pair: tuple[Token, Token]
    amount: int
    quote: Quote
    risk_analysis: RiskAnalysis


@dataclasses.dataclass(frozen=True, slots=True)
class ApprovalFailed:
    token: Token
    amount: int
    owner: AccountID
    spender: AccountID
    exception: Exception


@dataclasses.dataclass(frozen=True, slots=True)
class SubmitOrderFailed:
    pair: tuple[Token, Token]
    amount: int
    place_order_transaction: SentTransaction
    data: dict[str, Any]
    exception: Exception


@dataclasses.dataclass(frozen=True, slots=True)
class SourceNotWithdrawn:
    pair: tuple[Token, Token]
    amount: int
    order: OrderResponse
    wait_time: timedelta


@dataclasses.dataclass(frozen=True, slots=True)
class DestinationNotReceived:
    pair: tuple[Token, Token]
    amount: int
    order: OrderResponse
    wait_time: timedelta


TradeError = (
    InsufficientSourceBalance
    | QuoteFailed
    | QuoteLiquidityUnavailable
    | OrderRejected
    | RiskManagerRejection
    | ApprovalFailed  # Technically this should be a TransactionError, but ApprovableToken.approve() doesn't return events
    | SubmitOrderFailed
    | SourceNotWithdrawn
    | DestinationNotReceived
)


@dataclasses.dataclass(frozen=True, slots=True)
class CreateTransactionFailed:
    exception: Exception
    data: dict[str, Any]


@dataclasses.dataclass(frozen=True, slots=True)
class BroadcastTransactionFailed:
    transaction: Transaction
    exception: Exception
    data: dict[str, Any]


@dataclasses.dataclass(frozen=True, slots=True)
class WaitForTransactionReceiptFailed:
    transaction: SentTransaction
    exception: Exception
    data: dict[str, Any]


TransactionError = (
    CreateTransactionFailed
    | BroadcastTransactionFailed
    | WaitForTransactionReceiptFailed
)


@dataclasses.dataclass(frozen=True, slots=True)
class Trade:
    pair: tuple[Token, Token]
    quote: Quote
    order: OrderResponse


TradeEvent = Trade | TradeError | TransactionError



from __future__ import annotations
import abc
from abc import ABC
import dataclasses
from typing import Any

from hexbytes import HexBytes
from solders.signature import Signature
from tronpy.async_tron import AsyncTransactionRet

from ..chain import Chain
from ..chain_client import ChainClient, EthereumClient, SolanaClient, TronClient


# Proxy for a chain-specific sent transaction
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SentTransaction[ChainType: Chain](ABC):
    @classmethod
    def from_id(
        cls, client: ChainClient[ChainType], id: str
    ) -> SentTransaction[ChainType]:
        match client:
            case EthereumClient():
                return ethereum.EthereumSentTransaction(
                    client=client,
                    transaction_hash=HexBytes(id),
                )

            case SolanaClient():
                return solana.SolanaSentTransaction(
                    client=client,
                    signature=Signature.from_string(id),
                )

            case TronClient():
                return tron.TronSentTransaction(
                    chain_=client.chain,
                    native=AsyncTransactionRet(
                        iterable=(("txid", id),), client=client.native
                    ),
                )

            case _:
                raise NotImplementedError(f"Unimplemented chain: {client.chain}")

    # TODO: This currently doesn't follow the convention in the rest of the codebase that the `.id` property should produce a CAIP identifier that is unique across chains. Instead we return the transaction hash or equivalent, which is not necessarily unique across chains.
    @property
    @abc.abstractmethod
    def id(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def chain(self) -> ChainType:
        pass

    # Confirms the transaction if necessary and returns the receipt/response
    @abc.abstractmethod
    async def wait_for_receipt(self, **kwargs) -> Any:
        pass


# Proxy for a chain-specific transaction
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class Transaction[
    ChainType: Chain,
](ABC):
    @abc.abstractmethod
    async def broadcast(self) -> SentTransaction[ChainType]:
        pass

    @property
    @abc.abstractmethod
    def chain(self) -> ChainType:
        pass


# Avoid a circular import
from . import ethereum, solana, tron  # noqa: E402



from __future__ import annotations
import dataclasses
import typing

from tronpy import Tron
from tronpy.keys import PrivateKey, PublicKey

from ..chain import TronChain
from ..utility import tron
from .account import Account, AccountID


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class TronAccountID[Chain: TronChain](AccountID[Chain]):
    address_: str

    @classmethod
    def from_address(cls, chain: Chain, address: str) -> TronAccountID[Chain]:
        assert Tron.is_base58check_address(address)
        return cls(chain=chain, address_=address)

    @property
    @typing.override
    def address(self) -> str:
        return self.address_

    @typing.override
    def encode_address(self) -> bytes:
        return tron.encode_address(self.address_)


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class TronAccount[Chain: TronChain](Account[Chain]):
    private_key_: PrivateKey

    @classmethod
    def from_private_key(cls, chain: Chain, private_key: str) -> TronAccount[Chain]:
        return cls(
            chain=chain,
            private_key_=typing.cast(PrivateKey, PrivateKey.fromhex(private_key)),
        )

    @property
    @typing.override
    def address(self) -> str:
        return typing.cast(
            PublicKey, self.private_key_.public_key
        ).to_base58check_address()

    @property
    @typing.override
    def private_key(self) -> str:
        return self.private_key_.hex()

    @typing.override
    def encode_address(self) -> bytes:
        return tron.encode_address(self.address)

    @typing.override
    def downcast(self) -> TronAccountID:
        return TronAccountID(chain=self.chain, address_=self.address)



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from .book import OrderDirection, OrderExpiration, OrderFunding
from .types import Chain


class LiquiditySource(str, Enum):
    market_maker = "market_maker"
    cctp_direct = "cctp_direct"
    unavailable = "unavailable"


class OrderData(BaseModel):
    contract_address: str
    order_direction: OrderDirection
    order_funding: OrderFunding
    order_expiration: OrderExpiration
    target_address: str
    filler_address: str


class Quote(BaseModel):
    wallet_address: str
    src_chain: Chain
    dst_chain: Chain
    src_amount: int
    dst_amount: int
    bond_amount: int
    bond_fee: int
    src_asset_address: str
    dst_asset_address: str
    bond_asset_address: str
    challenge_offset: int
    challenge_window: int
    invalid_amount: bool
    liquidity_source: LiquiditySource
    created_at: str
    expires_at: str
    order_data: OrderData
    destination_transaction_fee: int
    destination_transaction_fee_usd: float
    destination_transaction_fee_after_subsidy: int
    destination_transaction_fee_after_subsidy_usd: float
    destination_transaction_fee_subsidy_usd: float
    reject_order: bool


class QuoteRequest(BaseModel):
    wallet_address: str
    target_address: Optional[str] = None
    src_chain: Chain
    dst_chain: Chain
    src_asset_address: str
    dst_asset_address: str
    src_amount: int



from __future__ import annotations
import dataclasses
import typing
from typing import Optional

from tronpy import AsyncContract, Tron
from tronpy.async_contract import AsyncContractMethod

from .. import config
from ..account import TronAccount, TronAccountID
from ..chain import TronChain
from ..chain_client import TronClient
from ..utility import tron
from ..transaction import TronSentTransaction, TronTransaction
from .token import ApprovableToken, NativeCoin


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class TronNativeCoin[Chain: TronChain](NativeCoin[Chain]):
    client: TronClient[Chain]

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @typing.override
    async def get_balance(self, account_id: TronAccountID[Chain]) -> int:
        return int(
            await self.client.native.get_account_balance(account_id.address) * 1_000_000
        )

    @typing.override
    async def transfer(
        self,
        sender: TronAccount[Chain],
        recipient: TronAccountID[Chain],
        amount: int,
    ) -> tuple[TronSentTransaction[Chain], dict]:
        builder = self.client.native.trx.transfer(
            sender.address, recipient.address, amount
        )

        transaction = await TronTransaction.from_builder(
            self.client, builder, sender, 0
        )

        sent_transaction = await transaction.broadcast()
        return await sent_transaction.wait_for_receipt()


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class TronToken[Chain: TronChain](ApprovableToken[Chain]):
    client: TronClient
    contract: AsyncContract

    @classmethod
    async def from_data(
        cls,
        client: TronClient[Chain],
        address: str,
        symbol: Optional[str],
        decimals: Optional[int],
    ) -> TronToken[Chain]:
        contract = await client.native.get_contract(address)
        assert contract.address
        contract.abi = config.trc20_abi

        if not symbol:
            symbol = typing.cast(str, await contract.functions.symbol())

        if not decimals:
            decimals = typing.cast(int, await contract.functions.decimals())

        return cls(client=client, contract=contract, symbol=symbol, decimals=decimals)

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @property
    @typing.override
    def asset_namespace(self) -> str:
        return "trc20"

    @property
    @typing.override
    def address(self) -> str:
        return self.contract.address  # type: ignore

    @typing.override
    def encode_address(self) -> bytes:
        return tron.encode_address(self.address)

    @typing.override
    async def get_balance(self, account_id: TronAccountID[Chain]) -> int:
        return await self.contract.functions.balanceOf(account_id.address)

    async def create_transfer_transaction(
        self,
        sender: TronAccount[Chain],
        recipient: TronAccountID[Chain],
        amount: int,
    ) -> TronTransaction[Chain]:
        return await TronTransaction.from_contract_method(
            self.client,
            sender,
            typing.cast(AsyncContractMethod, self.contract.functions.transfer),
            recipient.address,
            amount,
        )

    @typing.override
    async def transfer(
        self,
        sender: TronAccount[Chain],
        recipient: TronAccountID[Chain],
        amount: int,
    ) -> tuple[TronSentTransaction[Chain], dict]:
        transaction = await self.create_transfer_transaction(sender, recipient, amount)
        sent_transaction = await transaction.broadcast()
        return sent_transaction, await sent_transaction.wait_for_receipt()

    @typing.override
    async def get_allowance(
        self, owner: TronAccountID[Chain], spender: TronAccountID[Chain]
    ) -> int:
        return await self.contract.functions.allowance(owner.address, spender.address)

    async def create_approval_transaction(
        self,
        owner: TronAccount[Chain],
        spender: TronAccountID[Chain],
        amount: int,
    ) -> TronTransaction[Chain]:
        return await TronTransaction.from_contract_method(
            self.client,
            owner,
            typing.cast(AsyncContractMethod, self.contract.functions.approve),
            spender.address,
            amount,
        )

    @typing.override
    async def approve(
        self,
        owner: TronAccount[Chain],
        spender: TronAccountID[Chain],
        amount: int,
    ) -> dict:
        transaction = await self.create_approval_transaction(owner, spender, amount)
        sent_transaction = await transaction.broadcast()
        return await sent_transaction.wait_for_receipt()



from __future__ import annotations
from enum import Enum
from typing import Optional


from ...chain import Chain as NativeChain
from ...chain.constants import SupportedChain, NAME_TO_CHAIN


class Chain(str, Enum):
    ethereum = "ethereum"
    bsc = "bsc"
    avalanche = "avalanche"
    polygon = "polygon"
    arbitrum = "arbitrum"
    optimism = "optimism"
    base = "base"
    solana = "solana"
    celo = "celo"
    blast = "blast"
    scroll = "scroll"
    mantle = "mantle"
    opbnb = "opbnb"
    mode = "mode"
    tron = "tron"
    unichain = "unichain"

    @classmethod
    def from_chain(cls, chain: NativeChain) -> Chain:
        return CLIENT_CHAIN_TO_MACH_CHAIN[chain]

    def to_chain(self) -> NativeChain:
        return MACH_CHAIN_TO_CLIENT_CHAIN[self]

    def try_to_chain(self) -> Optional[NativeChain]:
        return MACH_CHAIN_TO_CLIENT_CHAIN.get(self)


MACH_CHAIN_TO_CLIENT_CHAIN = {
    (Chain(name.lower()) if chain != SupportedChain.BSC.value else Chain.bsc): chain
    for name, chain in NAME_TO_CHAIN.items()
}

CLIENT_CHAIN_TO_MACH_CHAIN = {
    chain: name for name, chain in MACH_CHAIN_TO_CLIENT_CHAIN.items()
}



from __future__ import annotations
import dataclasses
import typing

from eth_account import Account as EthAccount
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from web3 import Web3

from ..chain import EthereumChain
from ..utility import ethereum
from .account import Account, AccountID


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class EthereumAccountID[Chain: EthereumChain](AccountID[Chain]):
    address_: ChecksumAddress

    @classmethod
    def from_address(cls, chain: Chain, address: str) -> EthereumAccountID[Chain]:
        assert Web3.is_checksum_address(address)
        return cls(chain=chain, address_=typing.cast(ChecksumAddress, address))

    @property
    @typing.override
    def address(self) -> ChecksumAddress:
        return self.address_

    @typing.override
    def encode_address(self) -> bytes:
        return ethereum.encode_address(self.address_)


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class EthereumAccount[Chain: EthereumChain](Account[Chain]):
    native: LocalAccount

    @classmethod
    def from_private_key(cls, chain: Chain, private_key: str) -> EthereumAccount[Chain]:
        return cls(chain=chain, native=EthAccount.from_key(private_key))

    @typing.override
    def encode_address(self) -> bytes:
        return ethereum.encode_address(self.address)

    @property
    @typing.override
    def address(self) -> ChecksumAddress:
        return self.native.address

    @property
    @typing.override
    def private_key(self) -> str:
        return self.native.key.hex()

    @typing.override
    def downcast(self) -> EthereumAccountID:
        return EthereumAccountID(chain=self.chain, address_=self.native.address)



# TODO: Use this instead https://github.com/trustwallet/wallet-core/blob/master/registry.json

from enum import Enum

from .chain import Chain
from .ethereum import EthereumChain
from .solana import SolanaChain
from .tron import TronChain


class SupportedChain(Enum):
    # BITCOIN = "bip122:000000000019d6689c085ae165831e93"

    ETHEREUM = EthereumChain(chain_id=1)
    OPTIMISM = EthereumChain(chain_id=10)
    BSC = EthereumChain(chain_id=56)
    UNICHAIN = EthereumChain(chain_id=130)
    POLYGON = EthereumChain(chain_id=137)
    OPBNB = EthereumChain(chain_id=204)
    MANTLE = EthereumChain(chain_id=5000)
    BASE = EthereumChain(chain_id=8453)
    MODE = EthereumChain(chain_id=34443)
    ARBITRUM = EthereumChain(chain_id=42161)
    CELO = EthereumChain(chain_id=42220)
    AVALANCHE_C_CHAIN = EthereumChain(chain_id=43114)
    BLAST = EthereumChain(chain_id=81457)
    SCROLL = EthereumChain(chain_id=534352)

    SOLANA = SolanaChain(genesis_block_hash="5eykt4UsFv8P8NJdTREpY1vzqKqZKvdp")

    TRON = TronChain(genesis_block_hash="27Lqcw")


CHAIN_ID_TO_CHAIN = {chain.value.id: chain for chain in SupportedChain}


CHAIN_TO_NAME: dict[Chain, str] = {
    # SupportedChain.BITCOIN.value: "Bitcoin",
    SupportedChain.ETHEREUM.value: "Ethereum",
    SupportedChain.OPTIMISM.value: "Optimism",
    SupportedChain.BSC.value: "BSC",
    SupportedChain.UNICHAIN.value: "Unichain",
    SupportedChain.POLYGON.value: "Polygon",
    SupportedChain.OPBNB.value: "opBNB",
    SupportedChain.MANTLE.value: "Mantle",
    SupportedChain.BASE.value: "Base",
    SupportedChain.MODE.value: "Mode",
    SupportedChain.ARBITRUM.value: "Arbitrum",
    SupportedChain.CELO.value: "Celo",
    SupportedChain.AVALANCHE_C_CHAIN.value: "Avalanche",
    SupportedChain.BLAST.value: "Blast",
    SupportedChain.SCROLL.value: "Scroll",
    SupportedChain.SOLANA.value: "Solana",
    SupportedChain.TRON.value: "Tron",
}

NAME_TO_CHAIN = {name: chain for chain, name in CHAIN_TO_NAME.items()}

# Filled in by MachClient during initialization
# We have to fetch them from the backend because different instances (production, staging, dev) use different LayerZero versions and thus have different IDs
CHAIN_TO_LAYERZERO_ID: dict[Chain, int] = {}

LAYERZERO_ID_TO_CHAIN = {
    layerzero_id: chain for chain, layerzero_id in CHAIN_TO_LAYERZERO_ID.items()
}

# Filled in by MachClient during initialization
CCTP_DOMAINS: dict[Chain, int] = {}



from .token_messenger import TokenMessenger, TokenMessengerJSON
from .remote_token_messenger import RemoteTokenMessenger, RemoteTokenMessengerJSON
from .token_minter import TokenMinter, TokenMinterJSON
from .token_pair import TokenPair, TokenPairJSON
from .local_token import LocalToken, LocalTokenJSON



from typing import Annotated

from eth_typing import ChecksumAddress
from pydantic import BaseModel
from pydantic.functional_validators import AfterValidator
from web3 import Web3

# TODO: This is only for Ethereum, might change if this endpoint starts supporting more chains
Wallet = Annotated[ChecksumAddress, AfterValidator(Web3.to_checksum_address)]


class LeaderboardEntry(BaseModel):
    wallet: Wallet
    total_user_points: int


Leaderboard = list[LeaderboardEntry]


class PointsBreakdown(BaseModel):
    volume: int
    daily: int


class GenerateReferralCodeResponse(BaseModel):
    status: str
    message: str



import dataclasses
import typing

from .chain import Chain


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class SolanaChain(Chain):
    genesis_block_hash: str

    @property
    @typing.override
    def namespace(self) -> str:
        return "solana"

    @property
    @typing.override
    def reference(self) -> str:
        return self.genesis_block_hash



from __future__ import annotations
import abc
from abc import ABC

from ..asset import Token
from ..chain import Chain, EthereumChain, SolanaChain, TronChain
from ..transaction import SentTransaction


class Scanner(ABC):
    __slots__ = tuple()

    @staticmethod
    def create(chain: Chain) -> Scanner:
        match chain:
            case EthereumChain():
                return EIP3091Scanner(chain)
            case SolanaChain():
                return SolanaFM()
            case TronChain():
                return TronScan()
            case _ as chain:
                raise NotImplementedError(f"Unimplemented chain: {chain}")

    @property
    @abc.abstractmethod
    def base_url(self) -> str:
        pass

    @abc.abstractmethod
    def address(self, address: str) -> str:
        pass

    @abc.abstractmethod
    def transaction(self, transaction: SentTransaction) -> str:
        pass

    @abc.abstractmethod
    def token(self, token: Token) -> str:
        pass


# Avoid a circular import
from .eip3091 import EIP3091Scanner  # noqa: E402
from .solana import SolanaFM  # noqa: E402
from .tron import TronScan  # noqa: E402



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class InitializeArgs(typing.TypedDict):
    params: types.initialize_params.InitializeParams


layout = borsh.CStruct("params" / types.initialize_params.InitializeParams.layout)


class InitializeAccounts(typing.TypedDict):
    payer: Pubkey
    upgrade_authority: Pubkey
    authority_pda: Pubkey
    token_messenger: Pubkey
    token_minter: Pubkey
    token_messenger_minter_program_data: Pubkey
    token_messenger_minter_program: Pubkey
    event_authority: Pubkey
    program: Pubkey


def initialize(
    args: InitializeArgs,
    accounts: InitializeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(
            pubkey=accounts["upgrade_authority"], is_signer=True, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["authority_pda"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_messenger"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["token_minter"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["token_messenger_minter_program_data"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["token_messenger_minter_program"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["event_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["program"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xaf\xafm\x1f\r\x98\x9b\xed"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import dataclasses
from decimal import Decimal
import typing
from typing import Optional

from metaplex_python import Metadata
from solders import system_program
from solders.account_decoder import UiTokenAmount
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.system_program import TransferParams
from solders.transaction_status import EncodedConfirmedTransactionWithStatusMeta
from spl.token import instructions as spl_instructions
from spl.token import constants
from spl.token.async_client import AsyncToken
from spl.token.core import AccountInfo
from spl.token.instructions import ApproveCheckedParams, TransferCheckedParams

from ..account import SolanaAccount, SolanaAccountID
from ..chain import SolanaChain
from ..chain_client import SolanaClient
from ..utility import solana
from ..transaction import SolanaSentTransaction, SolanaTransaction
from .token import ApprovableToken, NativeCoin


DUMMY_PAYER = Keypair()


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class SolanaNativeCoin[Chain: SolanaChain](NativeCoin[Chain]):
    client: SolanaClient[Chain]

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @typing.override
    async def get_balance(self, account_id: SolanaAccountID[Chain]) -> int:
        response = await self.client.native.get_balance(account_id.pubkey)
        return response.value

    @typing.override
    async def transfer(
        self,
        sender: SolanaAccount[Chain],
        recipient: SolanaAccountID[Chain],
        amount: int,
    ) -> tuple[SolanaSentTransaction[Chain], Optional[EncodedConfirmedTransactionWithStatusMeta]]:
        transfer_instruction = system_program.transfer(
            TransferParams(
                from_pubkey=sender.keypair.pubkey(),
                to_pubkey=recipient.pubkey,
                lamports=amount,
            )
        )

        transaction = await SolanaTransaction.create(
            self.client,
            (transfer_instruction,),
            (sender,),
        )

        sent_transaction = await transaction.broadcast()
        return sent_transaction, await sent_transaction.wait_for_receipt()


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class SolanaToken[Chain: SolanaChain](ApprovableToken[Chain]):
    """
    TODO: I've avoided using AsyncToken as the native representation because it has some ugly behavior for transactions (approvals and transfers):
    - It requires a payer to be set on initialization for some reason.
        - This means for token transactions, you have to do `from copy import copy; token = copy(self.native); token.payer = <acutal payer>; token.transfer(...)`
    - It uses unversioned messages
    - It sends the transaction for you, so you can't use the SolanaTransaction class
    """

    client: SolanaClient[Chain]
    mint: Pubkey
    token_program_id: Pubkey

    @staticmethod
    def make_token(
        client: SolanaClient,
        token_id: Pubkey,
        token_program_id: Pubkey,
        payer: Keypair = DUMMY_PAYER,
    ) -> AsyncToken:
        return AsyncToken(
            conn=client.native,
            pubkey=token_id,
            program_id=token_program_id,
            payer=payer,
        )

    @classmethod
    async def from_data(
        cls,
        client: SolanaClient[Chain],
        address: str,
        symbol: Optional[str],
        decimals: Optional[int],
    ) -> SolanaToken[Chain]:
        mint = Pubkey.from_string(address)

        assert (
            token_account_info_response := await client.native.get_account_info(mint)
        ) and (
            token_account_info := token_account_info_response.value
        ), f"Unable to get token account info for {mint}"

        assert (token_program_id := token_account_info.owner) in (
            constants.TOKEN_PROGRAM_ID,
            constants.TOKEN_2022_PROGRAM_ID,
        ), f"{token_program_id} is not a valid token program ID"

        if not symbol:
            pda_str, _bump = Metadata.find_pda(str(mint))
            pda = Pubkey.from_string(pda_str)
            response = await client.native.get_account_info(pda)
            assert (value := response.value)
            metadata = Metadata(value.data)
            symbol = metadata.symbol()

        if not decimals:
            token = cls.make_token(client, mint, token_program_id)
            mint_info = await token.get_mint_info()
            decimals = mint_info.decimals

        return cls(
            client=client,
            mint=mint,
            token_program_id=token_program_id,
            symbol=symbol,
            decimals=decimals,
        )

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    # https://namespaces.chainagnostic.org/solana/caip19
    @property
    @typing.override
    def asset_namespace(self) -> str:
        return "token"

    @property
    @typing.override
    def address(self) -> str:
        return str(self.mint)
    
    @typing.override
    def encode_address(self) -> bytes:
        return solana.encode_address(self.mint)

    def associated_token_account(self, account_id: Pubkey) -> Pubkey:
        return spl_instructions.get_associated_token_address(
            owner=account_id,
            mint=self.mint,
            token_program_id=self.token_program_id,
        )

    async def get_raw_balance(
        self, account_id: SolanaAccountID[Chain]
    ) -> UiTokenAmount:
        token_account = self.associated_token_account(account_id.pubkey)
        response = await self.client.native.get_token_account_balance(token_account)

        return response.value

    @typing.override
    async def get_balance(self, account_id: SolanaAccountID[Chain]) -> int:
        raw_balance = await self.get_raw_balance(account_id)
        return int(raw_balance.amount)

    @typing.override
    async def get_balance_in_coins(self, account_id: SolanaAccountID[Chain]) -> Decimal:
        raw_balance = await self.get_raw_balance(account_id)
        return Decimal(raw_balance.ui_amount_string)

    async def get_token_account_info(
        self, account_id: SolanaAccountID[Chain]
    ) -> AccountInfo:
        # TODO: Wanted to avoid using AsyncToken for aforementioned reasons, but method that deserializes the AccountInfo from AsyncClient.get_account_info response is private
        token = self.make_token(self.client, self.mint, self.token_program_id)
        token_account = self.associated_token_account(account_id.pubkey)
        token_account_info = await token.get_account_info(token_account)

        return token_account_info

    async def token_account_exists(self, account_id: SolanaAccountID[Chain]) -> bool:
        token_account_info = await self.get_token_account_info(account_id)
        return token_account_info.is_initialized

    @typing.override
    async def transfer(
        self,
        sender: SolanaAccount[Chain],
        recipient: SolanaAccountID[Chain],
        amount: int,
    ) -> tuple[SolanaSentTransaction[Chain], Optional[EncodedConfirmedTransactionWithStatusMeta]]:
        source = self.associated_token_account(sender.keypair.pubkey())
        dest = self.associated_token_account(recipient.pubkey)

        transfer_checked_instruction = spl_instructions.transfer_checked(
            TransferCheckedParams(
                program_id=self.token_program_id,
                source=source,
                mint=self.mint,
                dest=dest,
                owner=sender.keypair.pubkey(),
                amount=amount,
                decimals=self.decimals,
            )
        )

        instructions = (
            (transfer_checked_instruction,)
            if await self.token_account_exists(recipient)
            else (
                # TODO: Using this to be safe but we could just use `create_associated_token_account()`
                spl_instructions.create_idempotent_associated_token_account(
                    payer=sender.keypair.pubkey(),
                    owner=recipient.pubkey,
                    mint=self.mint,
                    token_program_id=self.token_program_id,
                ),
                transfer_checked_instruction,
            )
        )

        transaction = await SolanaTransaction.create(
            self.client,
            instructions,
            (sender,),
        )

        sent_transaction = await transaction.broadcast()

        return sent_transaction, await sent_transaction.wait_for_receipt()

    @typing.override
    async def get_allowance(
        self,
        owner: SolanaAccountID[Chain],
        spender: SolanaAccountID[Chain],
    ) -> int:
        account_info = await self.get_token_account_info(owner)

        return (
            account_info.delegated_amount
            if account_info.delegate == spender.pubkey
            else 0
        )

    @typing.override
    async def approve(
        self,
        owner: SolanaAccount[Chain],
        spender: SolanaAccountID[Chain],
        amount: int,
    ) -> Optional[EncodedConfirmedTransactionWithStatusMeta]:
        source = self.associated_token_account(owner.keypair.pubkey())

        instructions = (
            spl_instructions.approve_checked(
                ApproveCheckedParams(
                    program_id=self.token_program_id,
                    source=source,
                    mint=self.mint,
                    delegate=spender.pubkey,
                    owner=owner.keypair.pubkey(),
                    amount=amount,
                    decimals=self.decimals,
                )
            ),
        )

        transaction = await SolanaTransaction.create(
            self.client,
            instructions,
            (owner,),
        )

        sent_transaction = await transaction.broadcast()

        return await sent_transaction.wait_for_receipt()



from __future__ import annotations
from decimal import Decimal
from typing import Annotated, Optional, TypedDict

from pydantic import AfterValidator, BaseModel

from ..types import MachChain


class ROI(TypedDict):
    times: Decimal
    currency: str
    percentage: Decimal


class AssetInfo(BaseModel):
    id: str
    symbol: Annotated[str, AfterValidator(lambda x: x.upper())]
    name: str
    image: str
    current_price: Decimal
    market_cap: int
    market_cap_rank: Optional[int]
    fully_diluted_valuation: Optional[Decimal]
    total_volume: Decimal
    high_24h: Optional[Decimal]
    low_24h: Optional[Decimal]
    price_change_24h: Optional[Decimal]
    price_change_percentage_24h: Optional[Decimal]
    market_cap_change_24h: Optional[Decimal]
    market_cap_change_percentage_24h: Optional[Decimal]
    circulating_supply: Decimal
    total_supply: Optional[Decimal]
    max_supply: Optional[Decimal]
    ath: Decimal
    ath_change_percentage: Decimal
    ath_date: str
    atl: Decimal
    atl_change_percentage: Decimal
    atl_date: str
    roi: Optional[ROI]
    last_updated: str
    decimals: int

    def __lt__(self, other: AssetInfo) -> bool:
        if not self.market_cap_rank:
            return True
        elif not other.market_cap_rank:
            return False

        return self.market_cap_rank < other.market_cap_rank


class AssetPricingData(BaseModel):
    chain: MachChain
    address: str
    symbol: str
    decimals: int
    price: Decimal
    daily_percent_change: float


class UserAssetData(BaseModel):
    chain: MachChain
    address: str
    balance: int
    symbol: Optional[str] = None
    price: Optional[Decimal]
    daily_percent_change: Optional[Decimal]



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class TokenMessengerJSON(typing.TypedDict):
    owner: str
    pending_owner: str
    local_message_transmitter: str
    message_body_version: int
    authority_bump: int


@dataclass
class TokenMessenger:
    discriminator: typing.ClassVar = b"\xa2\x04\xf24\x93\xf3\xdd`"
    layout: typing.ClassVar = borsh.CStruct(
        "owner" / BorshPubkey,
        "pending_owner" / BorshPubkey,
        "local_message_transmitter" / BorshPubkey,
        "message_body_version" / borsh.U32,
        "authority_bump" / borsh.U8,
    )
    owner: Pubkey
    pending_owner: Pubkey
    local_message_transmitter: Pubkey
    message_body_version: int
    authority_bump: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["TokenMessenger"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["TokenMessenger"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["TokenMessenger"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "TokenMessenger":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = TokenMessenger.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            owner=dec.owner,
            pending_owner=dec.pending_owner,
            local_message_transmitter=dec.local_message_transmitter,
            message_body_version=dec.message_body_version,
            authority_bump=dec.authority_bump,
        )

    def to_json(self) -> TokenMessengerJSON:
        return {
            "owner": str(self.owner),
            "pending_owner": str(self.pending_owner),
            "local_message_transmitter": str(self.local_message_transmitter),
            "message_body_version": self.message_body_version,
            "authority_bump": self.authority_bump,
        }

    @classmethod
    def from_json(cls, obj: TokenMessengerJSON) -> "TokenMessenger":
        return cls(
            owner=Pubkey.from_string(obj["owner"]),
            pending_owner=Pubkey.from_string(obj["pending_owner"]),
            local_message_transmitter=Pubkey.from_string(
                obj["local_message_transmitter"]
            ),
            message_body_version=obj["message_body_version"],
            authority_bump=obj["authority_bump"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class LzReceiveParamsJSON(typing.TypedDict):
    src_eid: int
    sender: list[int]
    nonce: int
    guid: list[int]
    message: list[int]
    extra_data: list[int]


@dataclass
class LzReceiveParams:
    layout: typing.ClassVar = borsh.CStruct(
        "src_eid" / borsh.U32,
        "sender" / borsh.U8[32],
        "nonce" / borsh.U64,
        "guid" / borsh.U8[32],
        "message" / borsh.Bytes,
        "extra_data" / borsh.Bytes,
    )
    src_eid: int
    sender: list[int]
    nonce: int
    guid: list[int]
    message: bytes
    extra_data: bytes

    @classmethod
    def from_decoded(cls, obj: Container) -> "LzReceiveParams":
        return cls(
            src_eid=obj.src_eid,
            sender=obj.sender,
            nonce=obj.nonce,
            guid=obj.guid,
            message=obj.message,
            extra_data=obj.extra_data,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "src_eid": self.src_eid,
            "sender": self.sender,
            "nonce": self.nonce,
            "guid": self.guid,
            "message": self.message,
            "extra_data": self.extra_data,
        }

    def to_json(self) -> LzReceiveParamsJSON:
        return {
            "src_eid": self.src_eid,
            "sender": self.sender,
            "nonce": self.nonce,
            "guid": self.guid,
            "message": list(self.message),
            "extra_data": list(self.extra_data),
        }

    @classmethod
    def from_json(cls, obj: LzReceiveParamsJSON) -> "LzReceiveParams":
        return cls(
            src_eid=obj["src_eid"],
            sender=obj["sender"],
            nonce=obj["nonce"],
            guid=obj["guid"],
            message=bytes(obj["message"]),
            extra_data=bytes(obj["extra_data"]),
        )



from __future__ import annotations
import dataclasses

from hdwallet import HDWallet
from hdwallet.const import PUBLIC_KEY_TYPES
from hdwallet.cryptocurrencies import Bitcoin, Ethereum, ICryptocurrency, Solana, Tron
from hdwallet.derivations import BIP44Derivation, CustomDerivation
from hdwallet.entropies import BIP39_ENTROPY_STRENGTHS, BIP39Entropy
from hdwallet.hds import BIP32HD, BIP44HD, IHD
from hdwallet.mnemonics import BIP39Mnemonic

from ..chain import Chain, EthereumChain, SolanaChain, TronChain
from .account import Account, AccountID


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class Wallet:
    mnemonic: str

    @staticmethod
    def create_base_hdwallet(
        cryptocurrency: type[ICryptocurrency] = Bitcoin,
        hd: type[IHD] = BIP44HD,
    ) -> HDWallet:
        return HDWallet(
            cryptocurrency=cryptocurrency,
            hd=hd,
            network=cryptocurrency.NETWORKS.MAINNET,
            public_key_type=PUBLIC_KEY_TYPES.COMPRESSED,
        )

    @classmethod
    def create(cls) -> Wallet:
        hdwallet = cls.create_base_hdwallet().from_entropy(
            entropy=BIP39Entropy(
                entropy=BIP39Entropy.generate(
                    strength=BIP39_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX
                )
            )
        )

        return cls(mnemonic=hdwallet.mnemonic())  # type: ignore

    # Check if the mnemonic is valid
    @classmethod
    def from_mnemonic(cls, mnemonic: str) -> Wallet:
        hdwallet = cls.create_base_hdwallet().from_mnemonic(
            mnemonic=BIP39Mnemonic(mnemonic=mnemonic)
        )
        return cls(mnemonic=hdwallet.mnemonic())  # type: ignore

    def create_hdwallet(
        self,
        cryptocurrency: type[ICryptocurrency] = Bitcoin,
        hd: type[IHD] = BIP44HD,
    ) -> HDWallet:
        return self.create_base_hdwallet(cryptocurrency, hd).from_mnemonic(
            mnemonic=BIP39Mnemonic(mnemonic=self.mnemonic)
        )

    @property
    def xprivate_key(self) -> str:
        return self.create_hdwallet().root_xprivate_key()  # type: ignore

    @property
    def xpublic_key(self) -> str:
        return self.create_hdwallet().root_xpublic_key()  # type: ignore

    # Derive the default address for the chain
    def derive_default(self, chain: Chain) -> HDWallet:
        match chain:
            case EthereumChain():
                cryptocurrency = Ethereum

                # Notice that we use the base coin type for Ethereum instead of the chain's coin type
                # So ie. BSC will use the same derivation path as Ethereum
                # This is to ensure that as many chains use the same keypair as possible
                return self.create_hdwallet(cryptocurrency).from_derivation(
                    derivation=BIP44Derivation(coin_type=cryptocurrency.COIN_TYPE)
                )

            case SolanaChain():
                cryptocurrency = Solana

                # https://github.com/talonlab/python-hdwallet/issues/109
                return self.create_hdwallet(cryptocurrency, BIP32HD).from_derivation(
                    derivation=CustomDerivation(
                        path=f"m/44'/{cryptocurrency.COIN_TYPE}'/0'/0'"
                    )
                )

            case TronChain():
                cryptocurrency = Tron

                return self.create_hdwallet(cryptocurrency).from_derivation(
                    derivation=BIP44Derivation(coin_type=cryptocurrency.COIN_TYPE)
                )

            case _:
                assert False, f"Unsupported chain: {chain} with type {type(chain)}"

    def account[ChainType: Chain](self, chain: ChainType) -> Account[ChainType]:
        hdwallet = self.derive_default(chain)
        return Account.from_hdwallet(chain, hdwallet)

    def account_id[ChainType: Chain](self, chain: ChainType) -> AccountID[ChainType]:
        hdwallet = self.derive_default(chain)
        return AccountID.from_hdwallet(chain, hdwallet)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class HandleReceiveMessageParamsJSON(typing.TypedDict):
    remote_domain: int
    sender: str
    message_body: list[int]
    authority_bump: int


@dataclass
class HandleReceiveMessageParams:
    layout: typing.ClassVar = borsh.CStruct(
        "remote_domain" / borsh.U32,
        "sender" / BorshPubkey,
        "message_body" / borsh.Bytes,
        "authority_bump" / borsh.U8,
    )
    remote_domain: int
    sender: Pubkey
    message_body: bytes
    authority_bump: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "HandleReceiveMessageParams":
        return cls(
            remote_domain=obj.remote_domain,
            sender=obj.sender,
            message_body=obj.message_body,
            authority_bump=obj.authority_bump,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "remote_domain": self.remote_domain,
            "sender": self.sender,
            "message_body": self.message_body,
            "authority_bump": self.authority_bump,
        }

    def to_json(self) -> HandleReceiveMessageParamsJSON:
        return {
            "remote_domain": self.remote_domain,
            "sender": str(self.sender),
            "message_body": list(self.message_body),
            "authority_bump": self.authority_bump,
        }

    @classmethod
    def from_json(
        cls, obj: HandleReceiveMessageParamsJSON
    ) -> "HandleReceiveMessageParams":
        return cls(
            remote_domain=obj["remote_domain"],
            sender=Pubkey.from_string(obj["sender"]),
            message_body=bytes(obj["message_body"]),
            authority_bump=obj["authority_bump"],
        )



# Reference:
# https://github.com/tristeroresearch/dapp/blob/dev/src/chains/solana/depositForBurn.ts

from __future__ import annotations
import typing

from solders.keypair import Keypair
from solders.pubkey import Pubkey

from ....account import SolanaAccount, SolanaAccountID
from ....asset import SolanaToken
from ....chain import SolanaChain
from ....chain_client import SolanaClient
from ....transaction import SolanaTransaction
from ...types import OrderData
from ....utility import solana as solana_utility
from .anchor import instructions
from .anchor.instructions import (
    DepositForBurnAccounts,
    DepositForBurnArgs,
)
from .anchor.types import DepositForBurnParams

if typing.TYPE_CHECKING:
    from ....client import MachClient


async def create_cctp_burn_transaction[SrcChainType: SolanaChain](
    *,
    client: MachClient,
    src_client: SolanaClient[SrcChainType],
    src_token: SolanaToken[SrcChainType],
    destination_domain: int,
    token_messenger_minter_address: str,
    order_data: OrderData,
    account: SolanaAccount[SrcChainType],
    recipient: bytes,
) -> SolanaTransaction[SrcChainType]:
    mint_recipient = Pubkey(recipient)

    params = DepositForBurnParams(
        amount=order_data.order_funding.src_amount_in,
        destination_domain=destination_domain,
        mint_recipient=mint_recipient,
    )

    args = DepositForBurnArgs(params=params)

    owner = typing.cast(SolanaAccountID, account).pubkey
    burn_token_account = typing.cast(SolanaToken, src_token).associated_token_account(
        owner
    )

    message_transmitter = Pubkey.from_string(
        client.cctp_message_transmitter_address(src_client.chain)
    )
    token_messenger_minter = Pubkey.from_string(token_messenger_minter_address)

    usdc_mint = typing.cast(SolanaToken, src_token).mint

    accounts = DepositForBurnAccounts(
        owner=owner,
        event_rent_payer=owner,
        sender_authority_pda=solana_utility.find_program_address(
            "sender_authority", token_messenger_minter
        ),
        burn_token_account=burn_token_account,
        message_transmitter=solana_utility.find_program_address(
            "message_transmitter", message_transmitter
        ),
        token_messenger=solana_utility.find_program_address(
            "token_messenger", token_messenger_minter
        ),
        remote_token_messenger=solana_utility.find_program_address(
            "remote_token_messenger",
            token_messenger_minter,
            (str(destination_domain),),
        ),
        token_minter=solana_utility.find_program_address(
            "token_minter", token_messenger_minter
        ),
        local_token=solana_utility.find_program_address(
            "local_token",
            token_messenger_minter,
            (usdc_mint,),
        ),
        burn_token_mint=usdc_mint,
        message_sent_event_data=Keypair().pubkey(),
        message_transmitter_program=message_transmitter,
        token_messenger_minter_program=token_messenger_minter,
        event_authority=solana_utility.find_program_address(
            "__event_authority", token_messenger_minter
        ),
        program=token_messenger_minter,
    )

    instruction = instructions.deposit_for_burn(args=args, accounts=accounts)

    return await SolanaTransaction.create(
        src_client, (instruction,), (typing.cast(SolanaAccount, account),)
    )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class AddRemoteTokenMessengerParamsJSON(typing.TypedDict):
    domain: int
    token_messenger: str


@dataclass
class AddRemoteTokenMessengerParams:
    layout: typing.ClassVar = borsh.CStruct(
        "domain" / borsh.U32, "token_messenger" / BorshPubkey
    )
    domain: int
    token_messenger: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "AddRemoteTokenMessengerParams":
        return cls(domain=obj.domain, token_messenger=obj.token_messenger)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"domain": self.domain, "token_messenger": self.token_messenger}

    def to_json(self) -> AddRemoteTokenMessengerParamsJSON:
        return {"domain": self.domain, "token_messenger": str(self.token_messenger)}

    @classmethod
    def from_json(
        cls, obj: AddRemoteTokenMessengerParamsJSON
    ) -> "AddRemoteTokenMessengerParams":
        return cls(
            domain=obj["domain"],
            token_messenger=Pubkey.from_string(obj["token_messenger"]),
        )



from eth_typing import ChecksumAddress

from ... import config
from ...account import AccountID, EthereumAccount
from ...chain import Chain, EthereumChain
from ...chain_client import EthereumClient
from ...transaction import EthereumTransaction
from ..types import OrderData


async def create_place_order_transaction[
    SrcChainType: EthereumChain,
](
    *,
    src_client: EthereumClient[SrcChainType],
    order_book_address: ChecksumAddress,
    order_data: OrderData,
    account: EthereumAccount[SrcChainType],
) -> EthereumTransaction[SrcChainType]:
    contract = src_client.w3.eth.contract(
        address=order_book_address,
        abi=config.ethereum_order_book_abi,
    )

    place_order_function = contract.functions.placeOrder(
        order_data.order_direction.to_eth(),
        order_data.order_funding.to_eth(),
        order_data.order_expiration.to_eth(),
        bytes.fromhex(order_data.target_address.removeprefix("0x")).rjust(32, b"\0"),
        order_data.filler_address,
    )

    return await EthereumTransaction.from_contract_function(
        src_client,
        place_order_function,
        account,
    )



import typing
from anchorpy.error import ProgramError


class InvalidAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Invalid Authority")

    code = 6000
    name = "InvalidAuthority"
    msg = "Invalid Authority"


class InvalidTokenOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "InvalidTokenOwner")

    code = 6001
    name = "InvalidTokenOwner"
    msg = "InvalidTokenOwner"


class InvalidTokenMintAddress(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "InvalidTokenMintAddress")

    code = 6002
    name = "InvalidTokenMintAddress"
    msg = "InvalidTokenMintAddress"


class InvalidTokenAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "InvalidTokenAmount")

    code = 6003
    name = "InvalidTokenAmount"
    msg = "InvalidTokenAmount"


class InvalidOwnerWithTradeMatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "InvalidOwnerWithTradeMatch")

    code = 6004
    name = "InvalidOwnerWithTradeMatch"
    msg = "InvalidOwnerWithTradeMatch"


class InvalidTradeMatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "InvalidTradeMatch")

    code = 6005
    name = "InvalidTradeMatch"
    msg = "InvalidTradeMatch"


class InvalidTokenStandard(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "InvalidTokenStandard")

    code = 6006
    name = "InvalidTokenStandard"
    msg = "InvalidTokenStandard"


class PayloadHashNotFound(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "PayloadHashNotFound")

    code = 6007
    name = "PayloadHashNotFound"
    msg = "PayloadHashNotFound"


class NotAgain(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "Already canceled or traded")

    code = 6008
    name = "NotAgain"
    msg = "Already canceled or traded"


class NotEvenStarted(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "Not even started")

    code = 6009
    name = "NotEvenStarted"
    msg = "Not even started"


class WrongMsgTypeError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "Wrong msg type")

    code = 6010
    name = "WrongMsgTypeError"
    msg = "Wrong msg type"


class WrongMsgDstIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "Can not find with dst index")

    code = 6011
    name = "WrongMsgDstIndex"
    msg = "Can not find with dst index"


class WrongMsgSrcToken(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "Can not find with src token address")

    code = 6012
    name = "WrongMsgSrcToken"
    msg = "Can not find with src token address"


class WrongMsgDstToken(ProgramError):
    def __init__(self) -> None:
        super().__init__(6013, "Can not find with dst token address")

    code = 6013
    name = "WrongMsgDstToken"
    msg = "Can not find with dst token address"


class WrongAuthorityToCancel(ProgramError):
    def __init__(self) -> None:
        super().__init__(6014, "Can not cancel with this authority")

    code = 6014
    name = "WrongAuthorityToCancel"
    msg = "Can not cancel with this authority"


class MinSellAmountConflict(ProgramError):
    def __init__(self) -> None:
        super().__init__(6015, "Min Sell Amount Conflict")

    code = 6015
    name = "MinSellAmountConflict"
    msg = "Min Sell Amount Conflict"


class InSufficientFundsOfOrder(ProgramError):
    def __init__(self) -> None:
        super().__init__(6016, "Insufficient Funds of Order")

    code = 6016
    name = "InSufficientFundsOfOrder"
    msg = "Insufficient Funds of Order"


class TokenNotApproved(ProgramError):
    def __init__(self) -> None:
        super().__init__(6017, "Token Not Approved")

    code = 6017
    name = "TokenNotApproved"
    msg = "Token Not Approved"


class InsufficientApproval(ProgramError):
    def __init__(self) -> None:
        super().__init__(6018, "Insufficient Approval")

    code = 6018
    name = "InsufficientApproval"
    msg = "Insufficient Approval"


class OrderNotValid(ProgramError):
    def __init__(self) -> None:
        super().__init__(6019, "Order Not Valid")

    code = 6019
    name = "OrderNotValid"
    msg = "Order Not Valid"


class OrderAlreadyMatched(ProgramError):
    def __init__(self) -> None:
        super().__init__(6020, "Order Already Matched")

    code = 6020
    name = "OrderAlreadyMatched"
    msg = "Order Already Matched"


class MatchAlreadyFinalized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6021, "Match Already Finalized")

    code = 6021
    name = "MatchAlreadyFinalized"
    msg = "Match Already Finalized"


class ArithmeticError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6022, "Arithmetic Error")

    code = 6022
    name = "ArithmeticError"
    msg = "Arithmetic Error"


class InsufficientOrderAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6023, "Insufficient Order Amount")

    code = 6023
    name = "InsufficientOrderAmount"
    msg = "Insufficient Order Amount"


class InvalidSendLibrary(ProgramError):
    def __init__(self) -> None:
        super().__init__(6024, None)

    code = 6024
    name = "InvalidSendLibrary"
    msg = None


class InvalidReceiveLibrary(ProgramError):
    def __init__(self) -> None:
        super().__init__(6025, None)

    code = 6025
    name = "InvalidReceiveLibrary"
    msg = None


class SameValue(ProgramError):
    def __init__(self) -> None:
        super().__init__(6026, None)

    code = 6026
    name = "SameValue"
    msg = None


class AccountNotFound(ProgramError):
    def __init__(self) -> None:
        super().__init__(6027, None)

    code = 6027
    name = "AccountNotFound"
    msg = None


class OnlySendLib(ProgramError):
    def __init__(self) -> None:
        super().__init__(6028, None)

    code = 6028
    name = "OnlySendLib"
    msg = None


class OnlyReceiveLib(ProgramError):
    def __init__(self) -> None:
        super().__init__(6029, None)

    code = 6029
    name = "OnlyReceiveLib"
    msg = None


class InvalidExpiry(ProgramError):
    def __init__(self) -> None:
        super().__init__(6030, None)

    code = 6030
    name = "InvalidExpiry"
    msg = None


class OnlyNonDefaultLib(ProgramError):
    def __init__(self) -> None:
        super().__init__(6031, None)

    code = 6031
    name = "OnlyNonDefaultLib"
    msg = None


class InvalidAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6032, None)

    code = 6032
    name = "InvalidAmount"
    msg = None


class InvalidNonce(ProgramError):
    def __init__(self) -> None:
        super().__init__(6033, None)

    code = 6033
    name = "InvalidNonce"
    msg = None


class Unauthorized(ProgramError):
    def __init__(self) -> None:
        super().__init__(6034, None)

    code = 6034
    name = "Unauthorized"
    msg = None


class ComposeNotFound(ProgramError):
    def __init__(self) -> None:
        super().__init__(6035, None)

    code = 6035
    name = "ComposeNotFound"
    msg = None


class InvalidPayloadHash(ProgramError):
    def __init__(self) -> None:
        super().__init__(6036, None)

    code = 6036
    name = "InvalidPayloadHash"
    msg = None


class LzTokenUnavailable(ProgramError):
    def __init__(self) -> None:
        super().__init__(6037, None)

    code = 6037
    name = "LzTokenUnavailable"
    msg = None


class ReadOnlyAccount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6038, None)

    code = 6038
    name = "ReadOnlyAccount"
    msg = None


class InvalidMessageLib(ProgramError):
    def __init__(self) -> None:
        super().__init__(6039, None)

    code = 6039
    name = "InvalidMessageLib"
    msg = None


class WritableAccountNotAllowed(ProgramError):
    def __init__(self) -> None:
        super().__init__(6040, None)

    code = 6040
    name = "WritableAccountNotAllowed"
    msg = None


CustomError = typing.Union[
    InvalidAuthority,
    InvalidTokenOwner,
    InvalidTokenMintAddress,
    InvalidTokenAmount,
    InvalidOwnerWithTradeMatch,
    InvalidTradeMatch,
    InvalidTokenStandard,
    PayloadHashNotFound,
    NotAgain,
    NotEvenStarted,
    WrongMsgTypeError,
    WrongMsgDstIndex,
    WrongMsgSrcToken,
    WrongMsgDstToken,
    WrongAuthorityToCancel,
    MinSellAmountConflict,
    InSufficientFundsOfOrder,
    TokenNotApproved,
    InsufficientApproval,
    OrderNotValid,
    OrderAlreadyMatched,
    MatchAlreadyFinalized,
    ArithmeticError,
    InsufficientOrderAmount,
    InvalidSendLibrary,
    InvalidReceiveLibrary,
    SameValue,
    AccountNotFound,
    OnlySendLib,
    OnlyReceiveLib,
    InvalidExpiry,
    OnlyNonDefaultLib,
    InvalidAmount,
    InvalidNonce,
    Unauthorized,
    ComposeNotFound,
    InvalidPayloadHash,
    LzTokenUnavailable,
    ReadOnlyAccount,
    InvalidMessageLib,
    WritableAccountNotAllowed,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: InvalidAuthority(),
    6001: InvalidTokenOwner(),
    6002: InvalidTokenMintAddress(),
    6003: InvalidTokenAmount(),
    6004: InvalidOwnerWithTradeMatch(),
    6005: InvalidTradeMatch(),
    6006: InvalidTokenStandard(),
    6007: PayloadHashNotFound(),
    6008: NotAgain(),
    6009: NotEvenStarted(),
    6010: WrongMsgTypeError(),
    6011: WrongMsgDstIndex(),
    6012: WrongMsgSrcToken(),
    6013: WrongMsgDstToken(),
    6014: WrongAuthorityToCancel(),
    6015: MinSellAmountConflict(),
    6016: InSufficientFundsOfOrder(),
    6017: TokenNotApproved(),
    6018: InsufficientApproval(),
    6019: OrderNotValid(),
    6020: OrderAlreadyMatched(),
    6021: MatchAlreadyFinalized(),
    6022: ArithmeticError(),
    6023: InsufficientOrderAmount(),
    6024: InvalidSendLibrary(),
    6025: InvalidReceiveLibrary(),
    6026: SameValue(),
    6027: AccountNotFound(),
    6028: OnlySendLib(),
    6029: OnlyReceiveLib(),
    6030: InvalidExpiry(),
    6031: OnlyNonDefaultLib(),
    6032: InvalidAmount(),
    6033: InvalidNonce(),
    6034: Unauthorized(),
    6035: ComposeNotFound(),
    6036: InvalidPayloadHash(),
    6037: LzTokenUnavailable(),
    6038: ReadOnlyAccount(),
    6039: InvalidMessageLib(),
    6040: WritableAccountNotAllowed(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from __future__ import annotations
import abc
from abc import ABC
import dataclasses
from typing import Any

import cachebox
from cachebox import Cache

from ..chain import Chain, EthereumChain, SolanaChain, TronChain


# Proxy for a chain-specific client
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class ChainClient[ChainType: Chain](ABC):
    chain: ChainType

    @staticmethod
    @cachebox.cached(Cache(0))
    async def _create(chain: ChainType) -> ChainClient[ChainType]:
        match chain:
            case EthereumChain():
                return await EthereumClient.create(chain)
            case SolanaChain():
                return await SolanaClient.create(chain)
            case TronChain():
                return await TronClient.create(chain)
            case _:
                raise NotImplementedError(f"Unsupported chain: {chain}")

    @classmethod
    @abc.abstractmethod
    async def create(cls, chain: ChainType) -> ChainClient[ChainType]:
        client = await cls._create(chain)

        if not await client.is_connected():
            await client.reconnect()
            assert await client.is_connected(), f"Failed to connect {chain} client"

        return client

    @abc.abstractmethod
    async def close(self) -> None:
        pass

    @abc.abstractmethod
    async def is_connected(self) -> bool:
        pass

    @abc.abstractmethod
    async def reconnect(self) -> None:
        pass

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ChainClient) and self.chain == other.chain

    def __hash__(self) -> int:
        return hash(self.chain)


# Avoid a circular import
from .ethereum import EthereumClient  # noqa: E402
from .solana import SolanaClient  # noqa: E402
from .tron import TronClient  # noqa: E402



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class AddLocalTokenParamsJSON(typing.TypedDict):
    pass


@dataclass
class AddLocalTokenParams:
    layout: typing.ClassVar = borsh.CStruct()

    @classmethod
    def from_decoded(cls, obj: Container) -> "AddLocalTokenParams":
        return cls()

    def to_encodable(self) -> dict[str, typing.Any]:
        return {}

    def to_json(self) -> AddLocalTokenParamsJSON:
        return {}

    @classmethod
    def from_json(cls, obj: AddLocalTokenParamsJSON) -> "AddLocalTokenParams":
        return cls()



from ..account import AccountBase
from ..asset import Token
from ..chain import Chain
from ..transaction import SentTransaction
from .scanner import Scanner


def address(chain: Chain, address: str) -> str:
    return Scanner.create(chain).address(address)


def account(account: AccountBase) -> str:
    return address(account.chain, account.address)


def transaction(transaction: SentTransaction) -> str:
    return Scanner.create(transaction.chain).transaction(transaction)


def token(token: Token) -> str:
    scanner = Scanner.create(token.chain)

    if token.is_native():
        return scanner.base_url

    return Scanner.create(token.chain).token(token)



import abc
from abc import ABC
from typing import Optional

from aiohttp import ClientSession
from pydantic import TypeAdapter

from ... import config
from ...account import AccountBase
from .. import utility
from .types import GenerateReferralCodeResponse, Leaderboard, PointsBreakdown


class MilesServerV2Mixin(ABC):
    __slots__ = ("__routes",)

    def __init__(self, url: str) -> None:
        # The double underscore forces name mangling to occur
        # This allows subclasses to have a different `self.routes` attribute
        self.__routes = config.config.miles_server.endpoints.add_url(url)

    @property
    @abc.abstractmethod
    def session(self) -> ClientSession:
        pass

    _leaderboard_validator = TypeAdapter(Leaderboard)

    async def get_points_leaderboard(self) -> Leaderboard:
        async with self.session.get(self.__routes.leaderboard) as response:
            bytes_result = await utility.to_bytes(response)

        return self._leaderboard_validator.validate_json(bytes_result)

    async def get_points_breakdown(self, account: AccountBase) -> PointsBreakdown:
        url = f"{self.__routes.breakdown}/{account.address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return PointsBreakdown.model_validate_json(bytes_result)

    _referral_code_validator = TypeAdapter(Optional[str])

    async def get_bound_referral_code(self, account: AccountBase) -> Optional[str]:
        url = f"{self.__routes.bound_referral_code}/{account.address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._referral_code_validator.validate_json(bytes_result)

    async def get_generated_referral_code(self, account: AccountBase) -> Optional[str]:
        url = f"{self.__routes.generated_referral_code}/{account.address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._referral_code_validator.validate_json(bytes_result)

    async def generate_referral_code(self, account: AccountBase) -> str:
        url = f"{self.__routes.generate_referral_code}/{account.address}"

        async with self.session.post(url) as response:
            bytes_result = await utility.to_bytes(response)

        return GenerateReferralCodeResponse.model_validate_json(bytes_result).message



from __future__ import annotations
import dataclasses
import typing
from typing import Optional, Sequence

from solana.rpc.types import TxOpts
from solders.address_lookup_table_account import AddressLookupTableAccount
from solders.transaction_status import EncodedConfirmedTransactionWithStatusMeta
from solders.instruction import Instruction
from solders.message import MessageV0
from solders.signature import Signature
from solders.transaction import VersionedTransaction

from ..account import SolanaAccount
from ..chain import SolanaChain
from ..chain_client import SolanaClient
from .transaction import SentTransaction, Transaction


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SolanaSentTransaction[Chain: SolanaChain](SentTransaction[Chain]):
    client: SolanaClient[Chain]
    signature: Signature

    @property
    @typing.override
    def id(self) -> str:
        return str(self.signature)

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @typing.override
    async def wait_for_receipt(
        self, **kwargs
    ) -> Optional[EncodedConfirmedTransactionWithStatusMeta]:
        await self.client.native.confirm_transaction(self.signature, **kwargs)
        response = await self.client.native.get_transaction(self.signature, max_supported_transaction_version=0)
        return response.value


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SolanaTransaction[Chain: SolanaChain](Transaction[Chain]):
    client: SolanaClient[Chain]
    transaction: VersionedTransaction

    @classmethod
    async def create(
        cls,
        client: SolanaClient[Chain],
        instructions: Sequence[Instruction],
        signers: Sequence[SolanaAccount[Chain]],
        address_lookup_table_accounts: Sequence[AddressLookupTableAccount] = tuple(),
    ) -> SolanaTransaction[Chain]:
        blockhash = await client.native.get_latest_blockhash()

        message = MessageV0.try_compile(
            payer=signers[0].keypair.pubkey(),
            instructions=instructions,
            address_lookup_table_accounts=address_lookup_table_accounts,
            recent_blockhash=blockhash.value.blockhash,
        )

        signer_keypairs = [signer.keypair for signer in signers]

        transaction = VersionedTransaction(message, signer_keypairs)

        return cls(client=client, transaction=transaction)

    @typing.override
    async def broadcast(
        self, opts: Optional[TxOpts] = None
    ) -> SolanaSentTransaction[Chain]:
        response = await self.client.native.send_transaction(self.transaction, opts)
        return SolanaSentTransaction(client=self.client, signature=response.value)

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class DepositForBurnParamsJSON(typing.TypedDict):
    amount: int
    destination_domain: int
    mint_recipient: str


@dataclass
class DepositForBurnParams:
    layout: typing.ClassVar = borsh.CStruct(
        "amount" / borsh.U64,
        "destination_domain" / borsh.U32,
        "mint_recipient" / BorshPubkey,
    )
    amount: int
    destination_domain: int
    mint_recipient: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "DepositForBurnParams":
        return cls(
            amount=obj.amount,
            destination_domain=obj.destination_domain,
            mint_recipient=obj.mint_recipient,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "amount": self.amount,
            "destination_domain": self.destination_domain,
            "mint_recipient": self.mint_recipient,
        }

    def to_json(self) -> DepositForBurnParamsJSON:
        return {
            "amount": self.amount,
            "destination_domain": self.destination_domain,
            "mint_recipient": str(self.mint_recipient),
        }

    @classmethod
    def from_json(cls, obj: DepositForBurnParamsJSON) -> "DepositForBurnParams":
        return cls(
            amount=obj["amount"],
            destination_domain=obj["destination_domain"],
            mint_recipient=Pubkey.from_string(obj["mint_recipient"]),
        )



from __future__ import annotations
import abc
from abc import ABC
import dataclasses
from typing import Any


# CAIP-2 Chain
# https://chainagnostic.org/CAIPs/caip-2
# https://chainagnostic.org/CAIPs/caip-288
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class Chain(ABC):
    @staticmethod
    def from_id(namespace: str, reference: str) -> Chain:
        match namespace:
            case "eip155":
                return EthereumChain(chain_id=int(reference))
            case "solana":
                return SolanaChain(genesis_block_hash=reference)
            case "tron":
                return TronChain(genesis_block_hash=reference)
            case _:
                assert False

    @classmethod
    def from_id_str(cls, id: str) -> Chain:
        namespace, reference = id.split(":")
        return cls.from_id(namespace, reference)

    @staticmethod
    def from_str(name: str) -> Chain:
        return constants.NAME_TO_CHAIN[name]

    @staticmethod
    def from_layerzero_id(id: int) -> Chain:
        return constants.LAYERZERO_ID_TO_CHAIN[id]

    @property
    @abc.abstractmethod
    def namespace(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def reference(self) -> str:
        pass

    @property
    def id(self) -> str:
        return f"{self.namespace}:{self.reference}"

    @property
    def layerzero_id(self) -> int:
        return constants.CHAIN_TO_LAYERZERO_ID[self]

    @property
    def cctp_domain(self) -> int:
        return constants.CCTP_DOMAINS[self]

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Chain) and self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __str__(self) -> str:
        return constants.CHAIN_TO_NAME[self]


# Avoid a circular import
from . import constants  # noqa: E402
from .ethereum import EthereumChain  # noqa: E402
from .solana import SolanaChain  # noqa: E402
from .tron import TronChain  # noqa: E402
